import sys
sys.path.append( '..' )
import stack_api.api_core as api_core
from src.comm.docker_ver import *
import typer
from pathlib import Path
from src.core.core import *

from src.comm.docker_ver import *
import os
path_home = os.getenv('LCP_DKR')+'/' if docker_ver() else str(Path.home())

app = typer.Typer()

# Checks if local files are installed
try:
    api = api_core.API()
    initialized = api.start_check()
    api.set_schema()
    api.commit('',False)
except:
    print(f'Stack has not been installed yet!')
    print(f'setting-up Stack...')
    api = api_core.API()
    initialized = api.start_check()
    try:
        api.set_schema()
    except:
        pass
    if initialized:
        api.commit('First commit')

### CLI Entry-points ###

@app.command("connect")
def connect_cli(uri: str):
    '''
        Connects the CLI to a remote dataset URI or directory
        Dataset can be located in the local system, an S3 bucket, or Google Cloud
    '''
    print('--------------------------------')
    print('- Connecting remote dataset... -')
    print('--------------------------------')
    
    # try:
    if docker_ver():
        assert(False)

    assert(not '.stack' in uri)

    if uri == '.':
        uri = str(os.path.abspath('.'))
    if uri[0:2] == '~/':
        uri = str(os.path.abspath('~/'))
    if uri[0:3] == 'C:/' or uri[0:3] == 'C:\\':
        uri[0:3] = '/c/'

    uri = uri.replace(path_home,'')

    api.init(uri)
    api.connect_post_cli()
    api.set_schema()
    api.reset_version()
    print('connection succesful')
    return {'success': True}
    # except:
        # print('connection unsuccesful!')

        # return {'success': False}

@app.command("put")
def put_command(file: str, target_subpath: str=''):
    '''
        Uploads a file to the dataset
    '''
    try:
        if len(target_subpath)>1:
            if target_subpath[-1] != '/':
                target_subpath = target_subpath + '/'
        api.upload_file_local_path(file, target_subpath)
        api.commit('')
        return True
    except:
        return Exception

@app.command("get")
def get_command(file):
    '''
        Downloads a file from the dataset
    '''
    try:
        return api.pull(file)
    except:
        return {}

@app.command("delete")
def delete_command(file: str):
    '''
        Deletes a file from the dataset
    '''
    api.remove(file)
    api.commit('', verbose=False)

@app.command("disconnect")
def disconnect_cli(uri: str):
    '''
        Disconnects the dataset from the CLI and removes it from
        the list of datasets tracked (does not delete diffs or history)
    '''
    try:
        api.init(uri)
        api.connect_post_api()
        api.set_schema()
        hierarchy = api.get_hierarchy()
        if len(hierarchy['children']) > 0:
            for child in hierarchy['children']:
                api.remove_child(child=child)
                try:
                    api.init(child)
                    api.connect_post_api()
                    api.set_schema()
                    api.remove_parent()
                except:
                    pass
            api.init(uri)
            api.connect_post_api()
            api.set_schema()
        if type(hierarchy['parent']) is str:
            if hierarchy['parent'] != '':
                assert(not '.stack' in hierarchy['parent'])
                if path_home in hierarchy['parent']:
                    hierarchy['parent'] = hierarchy['parent'].replace(path_home,'')
                try:
                    api.init(hierarchy['parent'])
                    api.connect_post_api()
                    api.set_schema()
                    api.remove_child(child=uri)
                except: 
                    pass    
        return {'success': api.disconnect_dataset(uri)}
    except:
        return {'success': False}

@app.command("slice")
def slice_cli(name: str, filter_json: str):
    '''
        Lists a slice of the filtered dataset and adds it to
        the list of slices
    '''
    try:
        filter = json.load(open(filter_json, "rb"))
        api.set_filters(filter)
        api.add_slice(slice_name=name)
        api.reset_filters()
        return {'success': True}
    except:
        return {'success': False}

@app.command("remove_slice")
def remove_slice_cli(name: str):
    '''
        Remove a slice from the list of slices
    '''
    try:
        api.remove_slice(slice_name=name)
        return {'success': True}
    except:
        return {'success': False}

@app.command("branch")
def branch_cli(uri: str, name: str = '', filter_json: str = '', type: str = 'copy'):
    '''
        Creates a branch of the dataset from the CLI and adds it to
        the list of datasets tracked
    '''
    try:
        if filter_json != '':
            filter = json.load(open(filter_json, "rb"))
            api.set_filters(filter)
        if name == '':
            name = uri
        parent = api.Initializer.storage.dataset
        api.set_hierarchy(child = name)
        api.branch(uri, type, name)
        api.set_hierarchy(parent = parent)
        if filter_json != '':
            api.reset_filters()
        return {'success': True}
    except:
        return {'success': False}

@app.command("merge")
def merge_cli(child: str = '', parent: str = ''):
    '''
        Merge a child branch of the dataset from the CLI and adds its changes to
        the a parent branch (us)
    '''
    try: 
        if child == '':
            if parent == '':
                child = api.Initializer.storage.dataset
                hierarchy = api.get_hierarchy()
                if type(hierarchy['parent']) is str:
                    if hierarchy['parent'] == '':
                        return {'success': False}
                    else:
                        api.merge(father=hierarchy['parent'], child=child)
                        api.commit(f"merged branch {child} to {hierarchy['parent']}")
                        return {'success': True}
            else:
                master = api.Initializer.storage.dataset
                api.merge(father=master, child=child)
                api.commit(f"merged branch {child} to {master}")
        else:
            if parent == '':
                hierarchy = api.get_hierarchy()
                if type(hierarchy['parent']) is str:
                    if hierarchy['parent'] == '':
                        return {'success': False}
                    else:
                        parent = hierarchy['parent']
            api.merge(father=parent, child=child)
            api.commit(f"merged branch {child} to {parent}")
        return {'success': True}
    except:
        return {'success': False}

@app.command("checkpoints")
def get_versions_command():
    try:
        versions = api.get_versions()
        for v in reversed(list(versions.keys())):
            if v != 'current_v':
                print(f"-- Version: {v} - Label: '{versions[v]['label']}' - Date: {versions[v]['date']} ")
    except:
        return {}

@app.command("add_checkpoint")
def add_version_command(name: str= ''):
    try:
        return {'success': api.add_version(name)}
    except:
        return {} 

@app.command('comment')
def add_tag_api(file: str, comment: str):
    api.add_tag(file, comment)
    return {'success': True}

@app.command('comments')
def get_tags_api(file: str):
    print(api.get_tags(file))
    return {'success': True}

@app.command('remove_comment')
def remove_tag_api(file: str, comment: str):
    api.remove_tag(file, comment)
    return {'success': True}

@app.command('remove_comments_all')
def remove_tags_api(file: str):
    api.selection_remove_all_tags(file)
    return {'success': True}

@app.command("delete_file_and_diffs")
def full_remove_key_api(file: str):
    '''
        Deletes a file from the dataset along with previous versions
    '''
    try:
        api_core.remove_full(file)
        print(f"Removed: {file}")
        api.commit('')
    except:
        print(f"Unable to remove: {file}")

@app.command("uri")
def uri_api():
    '''
        Gets the URI of the current dataset connected
    '''
    uri = api.config
    print('Current dataset: ' + uri['dataset'])
    print('Current storage: ' + uri['storage'])

@app.command("history")
def history_api():
    '''
        Prints a summary of the commits
    '''
    print_history(api.Initializer)

@app.command("datasets")
def dataset_api():
    '''
        Lists all the remote datasets that can connect to our CLI 
    '''
    api.print_datasets()

@app.command("status")
def status_api():
    '''
        Prints the status of the dataset
    '''
    try:
        status = api.status()
        if len(status['keys']) > 0:
            print('\n-------------------------')
            print('- Status of the dataset -')
            print('-------------------------\n')
            print('List of files in dataset:')
            for i in range(len(status['keys'])):
                print('\t-- '+status['keys'][i] + '\tlast modified: '+str(status['lm'][i]))
        else:
            print('everything is ok!')
            print('please add or commit a file')
    except:
        return Exception

@app.command("sync")
def sync_api(comment: str=''):
    '''
        Synchronizes new changes done in the connected dataset
    '''
    return {'success': api.commit(comment)}

@app.command("diff_versions")
def diff_api(version_a: str, version_b: str, file: str=''):
    '''
        List the changes applied to the dataset between version N and 
        version M (optional: give it a specific file to compare)
    '''
    if int(version_b) > int(version_a):
        tmp = version_a
        version_a = version_b
        version_b = tmp

    print_diff(api.Initializer, version_a, version_b, file)
    return True

@app.command("diff_csv")
def diff_csv(v1: str='current', v2: str='current', file: str=''):
    '''
        Printes the differences between two csv files in the path
    '''
    diff_metadata = api.load_csv_diff_metadata(key, v1, v2)
    print(json.dumps(diff_metadata, indent=4))
    diff = api.load_csv_diff(key, v1, v2)
    print(json.dumps(diff, indent=4))
    return True


@app.command("revert_to_version")
def revert_api(version):
    '''
        reverts the whole dataset to a specific version
    '''
    assert(version != '')
    revert_commit(api.Initializer, int(version))
    api.commit('reverted to version '+ version, verbose=False)
    return True

if __name__ == "__main__":
    app()



import sys
sys.path.append( '../../..' )
from datetime import *
from tzlocal import get_localzone
import cv2
import numpy as np
import time
import os
import copy
import io
import json
from pathlib import Path
from src.comm.docker_ver import *
path_home = os.getenv('LCP_DKR')+'/' if docker_ver() else str(Path.home())


from multiprocessing import Pool

def run_multiprocessing(func, i, n_processors):
    with Pool(processes=n_processors) as pool:
        return pool.map(func, i)

class yolo_schema(object):
	def __init__(self, init):
		self.init = init
		self.schema = None
		self.metadata = None
		self.schema_path = self.init.prefix_meta + 'schema.json'
		self.meta_path = self.init.prefix_meta + 'yolo_data.json'
		self.status = {}
		self.filtered = False
		self.sliced = False
		self.selected_slices = False
		self.bounding_box_thumbs = True
		
		self.in_version = False
		self.version_keys = None
		self.selected_version = None
		self.version_schema = ''

		ls, _ = self.init.storage.load_dataset_list()
		self.ls = ls
		self.ls_set = set(self.ls)

	def get_schema_from_key(self, key):
		dp = {}
		labels = self.get_labels(key)
		dp['key'] = key
		dp['lm'] = ''
		dp['classes'] = [labels[label]['0'] for label in labels]
		dp['labels'] = labels
		dp['n_classes'] = len(dp['classes'])
		dp['tags'] = []
		dp['resolution'] = self.get_resolution(key)
		dp['size'] = self.init.storage.get_size_of_file_global(key)/1024
		return dp

	def create_schema_file_in_parallel(self):
		schema = {}
		current = self.init.load_current()

		# t0 = time.time()
		with Pool(processes=6) as pool:
			arr_res =  pool.map(self.get_schema_from_key, current['keys'])
		
		schema['len'] = len(arr_res)
		# print(f'time to compute parallelized {time.time()-t0}s')

		return True

	def create_schema_file(self):
		# generates the schema file for a yolo dataset
		schema = {}
		current = self.init.load_current()
		k = 0
		idx = 0

		lm = current['lm']

		# TODO: Parallelize
		# finds the images
		
		for key in current['keys']:
			if self.is_image(key):
				labels = self.get_labels(key)
				dp = {}
				dp['tags'] = []
				dp['slices'] = []
				dp['key'] = key
				dp['lm'] = lm[idx]
				dp['classes'] = [labels[label]['0'] for label in labels]
				dp['labels'] = labels
				dp['n_classes'] = len(labels)
				dp['resolution'] = self.get_resolution(key)
				dp['size'] = self.init.storage.get_size_of_file_global(key)/1024
				schema[str(k)] = dp
				k += 1
			idx += 1

		schema['len'] = k

		# stores dp
		self.schema = schema
		self.init.storage.add_file_from_binary_global(self.schema_path,io.BytesIO(json.dumps(schema).encode('ascii')))
		self.init.storage.reset_buffer()
		self.copy_schema_to_latest_version_checkpoint()

		return True

	def copy_schema_to_latest_version_checkpoint(self):
		try:
			versions = self.init.load_versions()
		except:
			self.init.setup_versions()
			versions = self.init.load_versions()

		schema_copy_path = self.init.prefix_versions + '/schemas/' + f'{len(versions.keys())}.json' 
		schema_copy = copy.deepcopy(self.schema)

		for dp in schema_copy:
			if type(schema_copy[dp]) is dict:
				ver = self.init.get_latest_diff_number(schema_copy[dp]['key'])
				schema_copy[dp]['key'] = self.init.prefix_diffs + schema_copy[dp]['key'] + '/' + str(ver).zfill(10)

		self.init.storage.add_file_from_binary_global(schema_copy_path,io.BytesIO(json.dumps(schema_copy).encode('ascii')))
		keys = list(versions.keys())
		versions[keys[-1]]['schema_path'] = schema_copy_path

		metapath = self.init.prefix_versions + 'versions.json'
		self.init.storage.add_file_from_binary_global(metapath,io.BytesIO(json.dumps(versions).encode('ascii')))
		return True

	def select_version(self, version, schema_file, keys):
		self.in_version = True
		self.selected_version = version
		self.version_schema = schema_file
		self.ls = keys.values()
		self.ls_set = set(self.ls)

	def reset_to_current_version(self):
		self.in_version = False
		self.version_keys = None
		self.selected_version = None
		self.schema = None
		self.filtered = None
		self.sliced = None

		ls, _ = self.init.storage.load_dataset_list()
		# DO NOT modify self.ls without also modifying ls_set
		self.ls = ls
		self.ls_set = set(self.ls)
		
		self.schema = json.load(self.init.storage.load_file_global(self.schema_path))

		return True

	def get_schema(self):
		if self.schema == None:
			try:
				if self.in_version:
					self.schema = json.load(self.init.storage.load_file_global(self.version_schema))
				else:
					self.schema = json.load(self.init.storage.load_file_global(self.schema_path))
			except:
				self.create_schema_file()
		return self.schema

	def compute_meta_data(self):
		schema = self.get_schema()
		
		classes = []
		resolutions = []
		size = []
		lm = []
		tags = []
		classes_per_image = []
		slices = []

		n_class = {}
		n_res = {}
		n_lm = {}
		n_size = {}
		n_tags = {}
		n_slices = {}
		# TODO: Parallelize
		
		for val in schema:
			if type(self.schema[val]) is dict:
				if not len(schema[val]['classes']) in classes_per_image:
					classes_per_image.append(len(schema[val]['classes']))
				
				if 'slice' in schema[val].keys():
					for sl in schema[val]['slices']:
						if not sl in slices:
							slices.append(sl)
							n_slices[sl] = 1
						else:
							n_slices[sl] += 1

				for cl in schema[val]['classes']:
					if not cl in classes:
						classes.append(cl)
						n_class[cl] = 1
					else:
						n_class[cl] += 1
				
				if not schema[val]['resolution'] in resolutions:
					resolutions.append(schema[val]['resolution'])
					n_res[schema[val]['resolution']] = 1
				else:
					n_res[schema[val]['resolution']] += 1
				
				if not schema[val]['lm'] in lm:
					lm.append(schema[val]['lm'])
					n_lm[schema[val]['lm']] = 1
				else:
					n_lm[schema[val]['lm']] += 1
				
				if not schema[val]['size'] in size:
					size.append(schema[val]['size'])
					n_size[schema[val]['size']] = 1
				else:
					n_size[schema[val]['size']] += 1

				if 'tags' in schema[val].keys():
					for tag in schema[val]['tags']:
						if not tag in tags:
							tags.append(tag)
							n_tags[tag] = 1
						else:
							n_tags[tag] += 1

		metadata = {'classes': classes, 'resolutions': resolutions, 'size': size, 'lm': lm, 'slices': slices, 'n_slices': n_slices, 'tags': tags, 'n_class': n_class, 'n_res': n_res, 'n_lm': n_lm, 'n_tags': n_tags, 'classes_per_image': classes_per_image}
		self.metadata = metadata

		if not self.in_version:
			self.init.storage.add_file_from_binary_global(self.meta_path,io.BytesIO(json.dumps(metadata).encode('ascii')))
			self.init.storage.reset_buffer()

		return True

	def get_thumbnail(self, filename):
		if self.bounding_box_thumbs:
			# loads image string
			t0 = time.time()
			img_str = self.init.storage.load_file_global(filename)
			print(f'time to load image from {self.init.storage.type} {time.time()-t0}s')
			
			# formats to cv2
			nparr = np.fromstring(img_str.read(), np.uint8)
			img = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
			# trans_mask = img[:,:,3] == 0
			#replace areas of transparency with white and not transparent
			# img[trans_mask] = [255, 255, 255, 255]

			shapes = np.zeros_like(img, np.uint8)

			dh, dw, _ = img.shape
			if self.in_version:
				filename = os.path.dirname(filename.replace(self.init.prefix_diffs,''))
			basename = os.path.splitext(os.path.basename(filename))[0]
			
			ls = self.ls
			matches = [match for match in ls if basename+'.txt' in match]
			fl = self.init.storage.load_file_global(matches[0])

			for dt in fl.readlines():

				# Split string to float
				res = dt.split()
				try:
					cl = int(res[0])
				except:
					cl = res[0]
				try:
					x = float(res[1])
					y = float(res[2])
					w = float(res[3])
					h = float(res[4])
				except:
					x = 0
					y = 0
					w = 0
					h = 0

				color_str = str(cl)[::-1] + "c" + str(cl)
				
				hsh = 0

				for char in color_str:
					hsh = ord(char) + (hsh << 5) - hsh

				colour = "#"
				colour += ('00' + hex((hsh >> 0) & 0xFF))[-2:].replace('x','0')
				colour += ('00' + hex((hsh >> 8) & 0xFF))[-2:].replace('x','0')
				colour += ('00' + hex((hsh >> 16) & 0xFF))[-2:].replace('x','0')
				
				colour = colour[1:]
				
				color = tuple(int(colour[i:i+2],16) for i in (4,2,0))
				
				l = int((x - w / 2) * dw)
				r = int((x + w / 2) * dw)
				t = int((y - h / 2) * dh)
				b = int((y + h / 2) * dh)
				
				if l < 0:
					l = 0
				if r > dw - 1:
					r = dw - 1
				if t < 0:
					t = 0
				if b > dh - 1:
					b = dh - 1

				cv2.rectangle(shapes, (l, t), (r, b), color, -1)
				cv2.rectangle(img, (l, t), (r, b), color, 1)

			mask = shapes.astype(bool)

			img[mask] = cv2.addWeighted(img, 0.5, shapes, 0.5, 0)[mask]

			# retval, buf = cv2.imencode(".webp", img, [cv2.IMWRITE_WEBP_QUALITY, 1])
			# string_res = io.BytesIO(buf.tostring())
			string_res = io.BytesIO(cv2.imencode('.jpg', img)[1].tostring())
			
			return string_res
		else: 
			return self.init.storage.load_file_global(filename)

	def update_schema_file(self,added=[],modified=[],removed=[]):
		# loads the existing schema file
		schema = self.get_schema()
		# finds the images

		idx = int(schema['len'])
		for key in added:
			if self.is_image(key):
				dp = {}
				labels = self.get_labels(key)
				
				dp['key'] = key
				dp['lm'] = self.init.storage.load_file_metadata_global(key)['last_modified']
				dp['classes'] = [labels[label]['0'] for label in labels]
				dp['labels'] = labels
				dp['n_classes'] = len(dp['classes'])
				dp['resolution'] = self.get_resolution(key)
				dp['tags'] = []
				dp['slices'] = []
				dp['size'] = self.init.storage.get_size_of_file_global(key)

				schema[str(idx)] = dp
				
				idx += 1

		for key in modified:
			if self.is_image(key) or self.has_image(key):
				if self.has_image(key):
					key_img = self.has_image(key, path=True)
				else:
					key_img = key
				dp = {}
				labels = self.get_labels(key_img)
				
				dp['key'] = key_img
				dp['lm'] = self.init.storage.load_file_metadata_global(key)['last_modified']
				dp['classes'] = [labels[label]['0'] for label in labels]
				dp['labels'] = labels
				dp['n_classes'] = len(dp['classes'])
				dp['tags'] = self.get_tags(key_img)
				dp['slices'] = self.get_slices_key(key_img)
				dp['resolution'] = self.get_resolution(key_img)
				dp['size'] = self.init.storage.get_size_of_file_global(key_img)

				ref = 0
				for k, v in schema.items():
					if type(v) is dict:
						if v['key'] == key_img:
							ref = k

				schema[ref] = dp

		for key in removed:
			if self.is_image(key):
				ref = []
				for k, v in schema.items():
					if type(v) is dict:
						if v['key'] == key:
							ref.append(k)
				for k in ref:
					del schema[k]
					idx -= 1
				copy_schema = {}
				idx_copy = 0
				for k, v in schema.items():
					if type(v) is dict:
						copy_schema[str(idx_copy)] = v
						idx_copy += 1

				schema = copy_schema
		schema['len'] = idx

		# stores dp
		self.init.storage.add_file_from_binary_global(self.schema_path,io.BytesIO(json.dumps(schema).encode('ascii')))
		self.init.storage.reset_buffer()
		self.schema = schema

		return self.compute_meta_data()
	
	def get_tags(self, key):
		keys = self.schema.keys()
		for val in keys:
			if type(self.schema[val]) is dict:
				if key in self.schema[val]['key']:
					if 'tags' in self.schema[val].keys():
						return self.schema[val]['tags']
		return []

	def add_tag(self, key, tag):
		for val in self.schema.keys():
			if type(self.schema[val]) is dict:
				if key == self.schema[val]['key']:
					if 'tags' in self.schema[val].keys():
						if not tag in self.schema[val]['tags']:
							self.schema[val]['tags'].append(tag)
					else:
						self.schema[val]['tags'] = []
						self.schema[val]['tags'].append(tag)

		# stores schema file
		if self.in_version:
			self.init.storage.add_file_from_binary_global(self.schema_path,io.BytesIO(json.dumps(self.version_schema).encode('ascii')))
		else:
			self.init.storage.add_file_from_binary_global(self.schema_path,io.BytesIO(json.dumps(self.schema).encode('ascii')))
		self.init.storage.reset_buffer()

		return self.compute_meta_data()

	def add_many_tag(self, keys, tag):
		for val in self.schema.keys():
			if type(self.schema[val]) is dict:
				if self.schema[val]['key'] in keys:
					if 'tags' in self.schema[val].keys():
						if not tag in self.schema[val]['tags']:
							self.schema[val]['tags'].append(tag)
					else:
						self.schema[val]['tags'] = []
						self.schema[val]['tags'].append(tag)
					keys.remove(self.schema[val]['key'])

		# stores schema file
		if self.in_version:
			self.init.storage.add_file_from_binary_global(self.schema_path,io.BytesIO(json.dumps(self.version_schema).encode('ascii')))
		else:
			self.init.storage.add_file_from_binary_global(self.schema_path,io.BytesIO(json.dumps(self.schema).encode('ascii')))
		self.init.storage.reset_buffer()

		return self.compute_meta_data()

	def remove_tag(self, key, tag):
		keys = self.schema.keys()
		for val in keys:
			if type(self.schema[val]) is dict:
				if key in self.schema[val]['key']:
					if 'tags' in self.schema[val].keys():
						if tag in self.schema[val]['tags']:
							self.schema[val]['tags'].remove(tag)

		# stores schema file
		if self.in_version:
			self.init.storage.add_file_from_binary_global(self.schema_path,io.BytesIO(json.dumps(self.version_schema).encode('ascii')))
		else:
			self.init.storage.add_file_from_binary_global(self.schema_path,io.BytesIO(json.dumps(self.schema).encode('ascii')))
		self.init.storage.reset_buffer()

		return self.compute_meta_data()

	def remove_all_tags(self, key):
		keys = self.schema.keys()
		for val in keys:
			if type(self.schema[val]) is dict:
				if key in self.schema[val]['key']:
					if 'tags' in self.schema[val].keys():
						self.schema[val]['tags'] = []

		# stores schema file
		if self.in_version:
			self.init.storage.add_file_from_binary_global(self.schema_path,io.BytesIO(json.dumps(self.version_schema).encode('ascii')))
		else:
			self.init.storage.add_file_from_binary_global(self.schema_path,io.BytesIO(json.dumps(self.schema).encode('ascii')))
		self.init.storage.reset_buffer()

		return self.compute_meta_data()

	def many_remove_all_tags(self, keys):
		for val in self.schema.keys():
			if type(self.schema[val]) is dict:
				if self.schema[val]['key'] in keys:
					if 'tags' in self.schema[val].keys():
						self.schema[val]['tags'] = []
					keys.remove(self.schema[val]['key'])

		# stores schema file
		if self.in_version:
			self.init.storage.add_file_from_binary_global(self.schema_path,io.BytesIO(json.dumps(self.version_schema).encode('ascii')))
		else:
			self.init.storage.add_file_from_binary_global(self.schema_path,io.BytesIO(json.dumps(self.schema).encode('ascii')))
		self.init.storage.reset_buffer()

		return self.compute_meta_data()

	def is_image(self, key):
		# checks the extension
		extension = os.path.splitext(os.path.basename(key))[1]
		if extension in ['.jpg','.jpeg','.png','.bmp','.gif','.tiff','.svg']:
			return True
		return False

	def has_image(self, key, path = False):
		basename = os.path.splitext(os.path.basename(key))[0]
		ls = self.ls
		matches = [match for match in ls if basename in match]

		for m in matches:
			if self.is_image(m):
				if path:
					return m
				else:
					return True
		return False		

	def get_resolution(self, key):
		# reads image
		if self.init.storage.type == 'local':
			# HACK: in local it is faster to just read with opencv and not load the full image
			im = cv2.imread(key) 
			return str(im.shape[1]) + 'x' + str(im.shape[0])
		else:
			img_str = self.init.storage.load_file_global(key)
			
			# formats to cv2
			nparr = np.fromstring(img_str.read(), np.uint8)
			im = cv2.imdecode(nparr, cv2.IMREAD_COLOR)

			# returns shape
			return str(im.shape[1]) + 'x' + str(im.shape[0])

	def label_name(self, class_number):
		return str(class_number)

	def get_labels(self, filename, version='current'):
		# reads the labels
		if version == 'current':
			if self.in_version:
				filename = os.path.dirname(filename.replace(self.init.prefix_diffs,''))
			basename = os.path.splitext(os.path.basename(filename))[0]

			matches = [match for match in self.ls_set if basename+'.txt' in match]
			try: 
				labels_str = self.init.storage.load_file_global(matches[0])
			except:
				return {}
		else:
			assert(int(version) > 0)
			if self.in_version:
				filename = os.path.dirname(filename.replace(self.init.prefix_diffs,''))
			basename = os.path.splitext(os.path.basename(filename))[0]
			ls = set(self.ls)
			
			matches = [match for match in ls if basename+'.txt' in match]
			
			path = self.init.prefix_diffs + matches[0] + '/' + str(int(version)).zfill(10)
			labels_str = self.init.storage.load_file_global(path)

		labels = {}
		i = 0
		for line in labels_str.readlines():
			labels[str(i)] = {}
			j = 0
			for x in line.split():
				try: 
					if j == 0:
						labels[str(i)][str(j)] = x.decode("utf-8")
					else:
						labels[str(i)][str(j)] = float(x)
				except:
					labels[str(i)][str(j)] = 0
				j += 1
			i = i + 1
		return labels

	def get_labels_filename(self, filename, version = 'current'):
		if version == 'current':
			if self.in_version:
				filename = os.path.dirname(filename.replace(self.init.prefix_diffs,''))
			basename = os.path.splitext(os.path.basename(filename))[0]
			ls = self.ls
			matches = [match for match in ls if basename+'.txt' in match]
			return matches[0]
		else:
			assert(int(version) > 0)
			basename = os.path.splitext(os.path.basename(filename))[0]
			ls = self.ls
			matches = [match for match in ls if basename+'.txt' in match]
			return self.init.prefix_diffs + matches[0] + '/' + str(int(version)).zfill(10)

	def set_labels(self, filename, labels_array):
		# reads the labels
		basename = os.path.splitext(os.path.basename(filename))[0]
		ls = self.ls
		matches = [match for match in ls if basename+'.txt' in match]

		labels_string = ''

		for i in range(len(labels_array)-1):
			if type(labels_array[str(i)]) is dict:
				cl = labels_array[str(i)]['0']
				w = labels_array[str(i)]['1']
				h = labels_array[str(i)]['2']
				x = labels_array[str(i)]['3']
				y = labels_array[str(i)]['4']
			else:
				cl = labels_array[str(i)][0]
				w = labels_array[str(i)][1]
				h = labels_array[str(i)][2]
				x = labels_array[str(i)][3]
				y = labels_array[str(i)][4]

			labels_string = labels_string + f'{cl} {w} {h} {x} {y}\n'

		if (len(matches) > 0):
			self.init.storage.add_file_from_binary_global(matches[0],io.BytesIO(labels_string.encode("utf-8")))
		else:
			string = basename+'.txt'
			self.init.storage.add_file_from_binary(string,io.BytesIO(labels_string.encode("utf-8")))

		return True
	
	def branch(self, branch_name, type_ ='copy'):
		dataset = self.init.storage.dataset
		if self.init.storage.type == 'local':
			branch_name = path_home + '/' + branch_name + '/'
		if type_ == 'copy':
			for f in self.status['keys']:
				basename = os.path.splitext(os.path.basename(f))[0]
				ls = self.ls
				matches = [match for match in ls if basename+'.txt' in match]
				
				self.init.storage.copy_file_global(f,branch_name+f.replace(dataset,''))
				self.init.storage.copy_file_global(matches[0],branch_name+matches[0].replace(dataset,''))
		else:
			for f in self.status['keys']:
				basename = os.path.splitext(os.path.basename(f))[0]
				ls = self.ls
				matches = [match for match in ls if basename+'.txt' in match]
				
				self.init.storage.copy_file_global(f,branch_name+f.replace(dataset,''))
				self.init.storage.copy_file_global(matches[0],branch_name+matches[0].replace(dataset,''))
				self.init.storage.remove_file_global(dataset+f)
				self.init.storage.remove_file_global(dataset+matches[0])
		return True

	def merge(self, goal):
		dataset = self.init.storage.dataset
		if self.init.storage.type == 'local':
			goal = goal.replace(path_home,'')
			goal = path_home + '/' + goal + '/'
		if self.filtered:
			status = self.status['keys']
		else:
			status = self.read_all_files()
		for f in status['keys']:
			basename = os.path.splitext(os.path.basename(f))[0]
			ls = self.ls
			matches = [match for match in ls if basename+'.txt' in match]
			self.init.storage.copy_file_global(f,goal+f.replace(dataset,''))
			self.init.storage.copy_file_global(matches[0],goal+matches[0].replace(dataset,''))
		return True

	def read_all_files(self):
		# queries the json
		schema = self.get_schema()
		status = {'keys': [], 'lm': [], 'dp': []}
		for dp in schema:
			if type(self.schema[dp]) is dict:
				status['keys'].append(schema[dp]['key'])
				status['lm'].append(schema[dp]['lm'])
				status['dp'].append(dp)
		self.status = status
		return status

	def get_metadata(self):
		if self.filtered or self.in_version:
			schema = self.get_schema()
		
			classes = []
			resolutions = []
			size = []
			lm = []
			tags = []
			classes_per_image = []
			slices = []

			n_class = {}
			n_res = {}
			n_lm = {}
			n_size = {}
			n_tags = {}
			n_slices = {}

			for val in self.status['dp']:
				if not len(schema[val]['classes']) in classes_per_image:
					classes_per_image.append(len(schema[val]['classes']))
				
				if 'slices' in schema[val].keys():
					for sl in schema[val]['slices']:
						if not sl in slices:
							slices.append(sl)
							n_slices[sl] = 1
						else:
							n_slices[sl] += 1

				for cl in schema[val]['classes']:
					if not cl in classes:
						classes.append(cl)
						n_class[cl] = 1
					else:
						n_class[cl] += 1
				
				if not schema[val]['resolution'] in resolutions:
					resolutions.append(schema[val]['resolution'])
					n_res[schema[val]['resolution']] = 1
				else:
					n_res[schema[val]['resolution']] += 1
				
				if not schema[val]['lm'] in lm:
					lm.append(schema[val]['lm'])
					n_lm[schema[val]['lm']] = 1
				else:
					n_lm[schema[val]['lm']] += 1
				
				if not schema[val]['size'] in size:
					size.append(schema[val]['size'])
					n_size[schema[val]['size']] = 1
				else:
					n_size[schema[val]['size']] += 1

				if 'tags' in schema[val].keys():
					for tag in schema[val]['tags']:
						if not tag in tags:
							tags.append(tag)
							n_tags[tag] = 1
						else:
							n_tags[tag] += 1
			return {'classes': classes, 'resolutions': resolutions, 'size': size, 'lm': lm, 'slices': slices, 'n_slices': n_slices, 'tags': tags, 'n_class': n_class, 'n_res': n_res, 'n_lm': n_lm, 'n_tags': n_tags, 'classes_per_image': classes_per_image}
		else:
			return json.load(self.init.storage.load_file_global(self.meta_path))

	def reset_filters(self):
		self.filtered = False
		return True
	
	def apply_filters(self, filters={}):
		schema = self.get_schema()
		status = {'keys': [], 'lm': [], 'dp': []}

		if len(filters) == 0:
			self.filtered = False
			return self.status

		for dp in schema:
			if type(self.schema[dp]) is dict:
				if self.sliced:
					if 'slices' in schema[dp].keys():
						pre_add = any([sl in schema[dp]['slices'] for sl in self.selected_slices])
					else:
						pre_add = False
				else:
					pre_add = True

				if pre_add:
					add_num =  []
					add_class =  []
					add_res =  []
					add_name =  []
					add_tag =  []
					add_date =  []
					add_box =  []

					for f in filters:
						for filt in filters[f]:
							if filt == 'num_classes':
								
								min_cl = int(filters[f]['num_classes'][0])
								max_cl = int(filters[f]['num_classes'][1])
								
								if (len(schema[dp]['classes']) <= max_cl) and (len(schema[dp]['classes']) >= min_cl):
									add_num.append(True)
								else:
									add_num.append(False)
							
							if filt == 'class':
								if filters[f]['class'] in schema[dp]['classes']:
									add_class.append(True)
								else:
									add_class.append(False)
									
							if filt == 'resolution':
								if filters[f]['resolution'] == schema[dp]['resolution']:
									add_res.append(True)
								else:
									add_res.append(False)

							if filt == 'name':
								if filters[f]['name'] in schema[dp]['key']:
									add_name.append(True)
								else:
									add_name.append(False)

							if filt == 'tag':
								if 'tags' in schema[dp]:
									if filters[f]['tag'] in schema[dp]['tags']:
										add_tag.append(True)
									else:
										add_tag.append(False)
								else:
									add_tag.append(False)

							if filt == 'date':
								d_min = datetime.strptime(filters[f]['date'][0], '%Y/%m-%d').date()
								d_max = datetime.strptime(filters[f]['date'][1], '%Y/%m-%d').date()
								date = datetime.strptime(schema[dp]['lm'], '%m/%d/%Y, %H:%M:%S').astimezone(get_localzone()).date()

								if date <= d_max and date >= d_min:
									add_date.append(True)
								else:
									add_date.append(False)

							if filt == 'box_area':

								a_min = min(float(filters[f]['box_area'][0])/100, float(filters[f]['box_area'][1])/100)
								a_max = max(float(filters[f]['box_area'][0])/100, float(filters[f]['box_area'][1])/100)

								if 'labels' in schema[dp]:
									if len(schema[dp]['labels']) == 0:
										add_box.append(False)

									for i in range(len(schema[dp]['labels'])):
										area = float(schema[dp]['labels'][str(i)]['3']) * float(schema[dp]['labels'][str(i)]['4'])
										if (area >= a_min) and (area <= a_max):
											add_box.append(True)
										else:
											add_box.append(False)
								else:
									add_box.append(False)
					
					if len(add_class) == 0:
						add_class = [True]
					if len(add_res) == 0:
						add_res = [True]
					if len(add_name) == 0:
						add_name = [True]
					if len(add_tag) == 0:
						add_tag = [True]
					if len(add_box) == 0:
						add_box = [True]
					if len(add_num) == 0:
						add_num = [True]
					if len(add_date) == 0:
						add_date = [True]
						
					add = all([any(add_class),any(add_res),any(add_name),any(add_tag),any(add_box),any(add_num),any(add_date)])
					if add:
						status['keys'].append(schema[dp]['key'])
						status['lm'].append(schema[dp]['lm'])
						status['dp'].append(dp)

		self.filtered = True
		self.status = status
		return status

	def add_slice(self, slice_name=''):
		status = set(self.status['keys'])
		for val in self.schema.keys():
			if type(self.schema[val]) is dict:
				if self.schema[val]['key'] in status:
					if 'slices' in self.schema[val].keys():
						if not slice_name in self.schema[val]['slices']:
							self.schema[val]['slices'].append(slice_name)
					else:
						self.schema[val]['slices'] = [slice_name]

		# stores schema file
		if self.in_version:
			self.init.storage.add_file_from_binary_global(self.schema_path,io.BytesIO(json.dumps(self.version_schema).encode('ascii')))
		else:
			self.init.storage.add_file_from_binary_global(self.schema_path,io.BytesIO(json.dumps(self.schema).encode('ascii')))
		self.init.storage.reset_buffer()

		return self.compute_meta_data()

	def remove_slice(self, slice_name=''):
		for val in self.schema.keys():
			if type(self.schema[val]) is dict:
				if 'slices' in self.schema[val].keys():
					if slice_name in self.schema[val]['slices']:
						self.schema[val]['slices'].remove(slice_name)

		# stores schema file
		if self.in_version:
			self.init.storage.add_file_from_binary_global(self.schema_path,io.BytesIO(json.dumps(self.version_schema).encode('ascii')))
		else:
			self.init.storage.add_file_from_binary_global(self.schema_path,io.BytesIO(json.dumps(self.schema).encode('ascii')))
		self.init.storage.reset_buffer()

		return self.compute_meta_data()

	def get_slices(self):
		schema = self.get_schema()
		n_slices = {}
		slices = []
		for val in schema:
			if type(self.schema[val]) is dict:
				if 'slices' in schema[val].keys():
						for sl in schema[val]['slices']:
							if not sl in slices:
								slices.append(sl)
								n_slices[sl] = 1
							else:
								n_slices[sl] += 1
		return n_slices

	def get_slices_key(self, key):
		keys = self.schema.keys()
		for val in keys:
			if type(self.schema[val]) is dict:
				if key in self.schema[val]['key']:
					if 'slices' in self.schema[val].keys():
						return self.schema[val]['slices']
		return []
		
	def select_slice(self, slices=[]):
		if len(slices) == 0:
			self.sliced = False
			self.selected_slices = []
			return self.status
		self.sliced = True
		self.selected_slices = slices
		return self.apply_filters({'slic': []})
	
	def get_status(self):
		return self.status
	


import sys
sys.path.append( '../../../' )
import os
import shutil
from pathlib import Path
import time
from src.comm.docker_ver import *
path_home = os.getenv('LCP_DKR') if docker_ver() else str(Path.home())

class Local(object):
	"""docstring for Storage"""
	def __init__(self):
		self.type = "local"
		self.dataset = "./"
		self.prefix_ignore = ''
		self.credentials = {}

	def create_dataset(self,location,verbose=False):
		# transforms to absolute path
		if path_home in location:
			location.replace(path_home,'')

		if location[0] == '~':
			if len(location) > 1:
				location = path_home+location[1:]
			else:
				location = path_home
		if location[0] == '/':
			location = path_home + location
		else:
			location = path_home + '/' + location

		if location[-1] != '/':
			location = location + '/'
			self.dataset = location

		if not os.path.exists(location):
			print('dataset directory does not exist')
			print('failed to connect')
			raise Exception('dataset directory does not exist')

		return False

	def check_if_empty(self,path):
		if os.path.exists(path):
			dir = os.listdir(path)
			return (len(dir) == 0)
		else:
			return True

	def add_file(self,filepath,target_name='',subpath=''):
		if filepath[-1] == '/':
			for root,dirs,files in os.walk(filepath):
				for file in files:
					if not os.path.exists(os.path.dirname(self.dataset+subpath)):
						os.makedirs(os.path.dirname(self.dataset+subpath))
					head_tail = os.path.split(os.path.join(root,file))
					shutil.copyfile(os.path.join(root,file), self.dataset+subpath+head_tail[1])	
				self.reset_buffer()
		else:
			if not os.path.exists(os.path.dirname(self.dataset+subpath)):
				os.makedirs(os.path.dirname(self.dataset+subpath))
			head_tail = os.path.split(filepath)
			shutil.copyfile(filepath, self.dataset+subpath+head_tail[1])
		return True

	def add_folder(self,folderpath):
		if not os.path.exists(self.dataset + folderpath):
			os.makedirs(self.dataset + folderpath)
		return True

	def add_file_from_binary(self,filename,f):
		return self.add_file_from_binary_global(self.dataset+filename,f)

	def add_file_from_binary_global(self,filename,f):
		if not os.path.exists(os.path.dirname(filename)):
			os.makedirs(os.path.dirname(filename))
		binary_file = open(filename, "wb")
		binary_file.write(f.read())
		binary_file.close()
		return True

	def remove_file(self,filename):
		return self.remove_file_global(self.dataset+filename)

	def remove_file_global(self,filename):
		if os.path.exists(filename):
			if filename[-1] == '/':
				shutil.rmtree(filename)
			else:
				os.remove(filename)
		return True

	def load_file(self,filename):
		return self.load_file_global(self.dataset+filename)

	def load_file_bytes(self,filename,bi,bf):
		return self.load_file_global(self.dataset+filename,bi,bf)

	def load_file_global(self,filename):
		return open(filename,'rb')

	def load_file_global_bytes(self,filename,bi,bf):
		f = open(filename, "rb")
		f.seek(bi)
		data = f.read(bf-bi)
		return data

	def load_file_metadata(self,filename):
		path = self.dataset+filename
		metadata = {
			'key' : self.dataset+filename,
			'date_loaded' : time.strftime("%m/%d/%Y, %H:%M:%S", time.gmtime(os.path.getmtime(path))),
			'date_added' : time.strftime("%m/%d/%Y, %H:%M:%S", time.gmtime(os.path.getctime(path))),
			'last_modified' : time.strftime("%m/%d/%Y, %H:%M:%S", time.gmtime(os.path.getmtime(path))),
		}
		return metadata

	def load_file_metadata_global(self,filename):
		path = filename
		metadata = {
			'key' : self.dataset+filename,
			'date_loaded' : time.strftime("%m/%d/%Y, %H:%M:%S", time.gmtime(os.path.getmtime(path))),
			'date_added' : time.strftime("%m/%d/%Y, %H:%M:%S", time.gmtime(os.path.getctime(path))),
			'last_modified' : time.strftime("%m/%d/%Y, %H:%M:%S", time.gmtime(os.path.getmtime(path))),
		}
		return metadata

	def load_dataset(self):
		file_m = []
		for root, dirs, files in os.walk(self.dataset, topdown=False):
			for name in files:
				path = os.path.join(root, name)
				if path[-1] != '/':
					metadata = {
						'key' : path,
						'last_modified' : time.strftime("%m/%d/%Y, %H:%M:%S", time.gmtime(os.path.getmtime(path)))
					}
					file_m.append(metadata)
		return file_m

	
	def load_dataset_list(self):
		return self.load_list_in_path(self.dataset)

	def load_list_in_path(self,path):
		keys = []
		last_m = []

		for root, dirs, files in os.walk(path, topdown=False):
			for name in files:
				keys.append(os.path.join(root, name))
				last_m.append(time.strftime("%m/%d/%Y, %H:%M:%S", time.gmtime(os.path.getmtime(os.path.join(root, name)))))
		return keys, last_m

	def load_files_in_dataset(self):
		files = []
		for root, dirs, files in os.walk(self.dataset, topdown=False):
			for name in files:
				path = os.path.join(root, name)
		return True

	def list_files_in_path(self,dir_path):
		for root, dirs, files in os.walk(self.dataset + dir_path, topdown=False):
			for name in files:
				path = os.path.join(root, name)
		return True

	def copy_file(self,filepath,full_target_name):
		return self.copy_file_global(self.dataset + filepath, self.dataset + full_target_name)

	def copy_file_global(self,filepath,full_target_name):
		head_tail = os.path.split(filepath)
		os.makedirs(os.path.dirname(full_target_name), exist_ok=True)
		if filepath[-1] != '/':
			shutil.copyfile(filepath, full_target_name)
		return True

	def get_size_of_file(self, filepath):
		return self.get_size_of_file_global(self.dataset + filepath)

	def get_size_of_file_global(self, filepath):
		return os.path.getsize(filepath)

	def reset_buffer(self):
		pass

def main():
	storage = Local()
	
	print('create dataset')
	storage.create_dataset('/Users/bernardo/test_data/')
	
	print('add folder')
	storage.add_folder('.stack')
	print('add subfolder')
	storage.add_folder('subtree')
	print('list in dataset')
	storage.load_files_in_dataset()
	print('list in path')
	storage.list_files_in_path('')
	
	print('adds a file')
	storage.add_file('../../tests/image.png')
	print(storage.load_file_metadata('image.png'))

	storage.copy_file('image.png',storage.dataset+'/copies/image4.png')
	storage.list_files_in_path('')

if __name__ == '__main__':
	main()


from operator import le
import sys
sys.path.append( '..' )
from src.core.init import Initializer
from src.core.core import *
from src.training.projects import projects

from src.storage.classes.s3 import S3Bucket
from src.storage.classes.gcs import GCSBucket
from src.storage.classes.local import Local

from src.dataset.object_detection.yolo import yolo_schema
from src.dataset.object_detection.labelbox import labelbox_schema

from src.dataset.named_entity_recognition.spacy import spacy_ner_schema

from pathlib import Path
import pickle
import os
from io import StringIO
from src.comm.docker_ver import *
path_home = os.getenv('LCP_DKR')+'/' if docker_ver() else str(Path.home())
import pandas as pd

class API(object):
    """docstring for CLI"""
    def __init__(self, reset=False, cli=False):
        super(API, self).__init__()
        self.key_bin = None
        self.Initializer = None
        self.config = None
        self.schema_class = None
        self.current = None
        
        self.in_version = False
        self.selected_version = None
        self.version_keys = None

        self.filtered = False
        self.projects = None
        self.filters = {}
        if not Path(path_home+'/datasets.stack').exists():
            self.set_datasets({})
        if reset:
            self.Initializer = None
        
        if not Path(path_home+'/.config_stack').exists():
            self.config = {}
            self.set_config()
        else:
            self.config = self.get_config()
            try:
                self.storage_name = self.config['storage']
                self.dataset_name = self.config['dataset']
                ctype = self.config['type']
                schema = self.config['schema']
            except:
                self.config = {'storage': '', 'schema': 'files'}
                self.set_config()
                ctype = ''
            try:
                if ctype == 'local':
                    cloud = Local()
                    cloud.create_dataset(self.config['dataset'])
                    self.Initializer = Initializer(cloud)
                    self.dataset_name = self.config['storage']
                    self.storage_name = self.config['dataset']
                    self.Initializer.schema = schema
                elif ctype == 's3':
                    cloud = S3Bucket(self.config['bucket'])
                    if cli:
                        cloud.connect_bucket()
                    else:
                        cloud.reconnect_bucket_api()
                    cloud.create_dataset(self.config['dataset'])
                    self.Initializer = Initializer(cloud)
                    self.Initializer.schema = schema
                elif ctype == 'gcs':
                    cloud = GCSBucket(self.config['bucket'])
                    if cli:
                        cloud.connect_bucket()
                    else:
                        cloud.reconnect_bucket_api()
                    cloud.create_dataset(self.config['dataset'])
                    self.Initializer = Initializer(cloud)
                    self.Initializer.schema = schema
                else:
                    self.Initializer = None
            except:
                self.Initializer = None
        
    def init(self, storage = None):
        # builds a config file       
        if Path(path_home+'/.config_stack').exists():
            config = {} 
            if storage == None:
                return False
            if 's3:/' in storage.lower():
                bucket_data = storage.split("/")[1:]
                config['bucket'] = bucket_data[1]
                config['dataset'] = ''.join(bucket_data[2:])+'/'
                config['type'] = 's3'
            elif 'gs:/' in storage.lower():
                bucket_data = storage.split("/")[1:]
                config['bucket'] = bucket_data[1]
                config['dataset'] = ''.join(bucket_data[2:])+'/'
                config['type'] = 'gcs'
            else:
                if path_home in storage:
                    storage.replace(path_home,'')
                if storage[0] == '/':
                    storage = storage[1:]
                if storage[-1] == '/':
                    storage = storage[:-1]
                config['dataset'] = storage
                config['type'] = 'local'
            config['storage'] = storage
            
            # stores the config file
            self.config = config
            self.filtered = False
            # creates dataset
            return True
        else:
            config = {'storage': ''}
            self.config = config
            return False

    def init_config(self):
        # builds a config file       
        if Path(path_home+'/.config_stack').exists():
            return True
        else:
            config = {'storage': '', 'schema': 'files'}
            self.config = config
            self.set_config()
            return False

    def set_gs_key(self, file):
        self.key_bin = file
        return True

    def get_config(self):
        file2 = open(path_home+'/.config_stack', 'rb')
        config = pickle.load(file2)
        file2.close()
        self.config = config
        return config

    def set_config(self):
        file = open(path_home+'/.config_stack', 'wb')
        pickle.dump(self.config,file)
        file.close()

    def get_datasets(self):
        file1 = open(path_home+'/datasets.stack', 'rb')
        datasets = pickle.load(file1)
        file1.close()
        return datasets

    def get_labels(self, filename, version='current'):
        
        if(self.config['schema'] != 'files'):
            return self.schema_class.get_labels(filename, version)

        return []

    def set_bounding_boxes(self, val = 'false'):
        if self.config['schema'] == 'yolo' or self.config['schema'] == 'labelbox':
            if val == 'true':
                self.schema_class.bounding_box_thumbs = False
            else:
                self.schema_class.bounding_box_thumbs = True
        return True

    def load_thumbnail(self, file):
        if self.config['schema'] == 'yolo' or self.config['schema'] == 'labelbox':
            try: 
                return self.schema_class.get_thumbnail(file)
            except:
                return self.load_file_binary(file, 'current')    
        else:
            return self.load_file_binary(file, 'current')

    def set_labels(self, data):
        if self.config['schema'] == 'yolo' or self.config['schema'] == 'labelbox':
            if self.in_version:
                filename =  os.path.dirname(data['keyId'].replace(self.Initializer.prefix_diffs,''))
                in_version = True
                version = self.selected_version
                self.reset_version()
            else:
                in_version = False
                filename = data['keyId']
            self.schema_class.set_labels(filename, data)

            if in_version:
                self.select_version(version)
        elif self.config['schema'] == 'spacy_ner':
            if self.in_version:
                self.reset_version()
            else:
                filename = data['keyId']
            self.schema_class.set_labels(filename, data)

            if in_version:
                self.select_version(version)
        return {'success': True}

    def set_datasets(self, datasets):
        file1 = open(path_home+'/datasets.stack', 'wb')
        pickle.dump(datasets,file1)
        file1.close()

    def print_datasets(self):
        datasets = self.get_datasets()

        if len(datasets.keys()) > 0:
            print('List of remote datasets:\n')
            for s in datasets.keys():
                print('-- '+datasets[s]['name']+' in '+datasets[s]['storage'])
            print('')
            print('run \'stack connect [dataset_uri]\' to connect one of these datasets')
        else:
            print('no datasets to show')
        return True
    
    def get_dataset_name(self,uri):
        if uri == '':
            return ''
        datasets = self.get_datasets()
        if uri[-1] == '/':
            uri = uri[:-1]
        if uri[0] == '/':
            uri = uri[1:]
        for s in datasets.keys():
            if uri == datasets[s]['storage']:
                return datasets[s]['name']
        return ''

    def get_dataset_URI(self,name):
        datasets = self.get_datasets()
        for s in datasets.keys():
            if name == datasets[s]['name']:
                return datasets[s]['storage']
        return ''
    
    def set_schema(self):

        if self.config['schema'] == 'yolo':
            self.schema_class = yolo_schema(self.Initializer)
            try:
                metapath = self.schema_class.schema_path
                self.schema_class.schema = json.load(self.Initializer.storage.load_file_global(metapath))
            except:
                print('creating schema file')
                self.schema_class.create_schema_file()
                print('computing metadata')
                self.schema_class.compute_meta_data()
        elif self.config['schema'] == 'labelbox':
            self.schema_class = labelbox_schema(self.Initializer)
            try:
                metapath = self.schema_class.meta_path
                json.load(self.Initializer.storage.load_file_global(metapath))
            except:
                self.schema_class.create_schema_file()
                self.schema_class.compute_meta_data()
        elif self.config['schema'] == 'spacy_ner':
            self.schema_class = spacy_ner_schema(self.Initializer)
            try:
                metapath = self.schema_class.meta_path
                json.load(self.Initializer.storage.load_file_global(metapath))
            except:
                self.schema_class.create_schema_file()
                self.schema_class.compute_meta_data()

    def reset_schema(self):
        self.reset_version()
        if self.config['schema'] != 'files':
            self.schema_class.create_schema_file()
            self.schema_class.compute_meta_data()

    def merge(self, father='', child=''):
        self.reset_version()
        if father[-1] != '/':
            father += '/'
        if child[-1] != '/':
            child += '/'

        if 's3://' in father:
            uri = 's3://'+self.Initializer.storage.BUCKET_NAME+'/'
            if uri in child:
                uri = ''
        elif 'gs://' in father:
            uri = 'gs://'+self.Initializer.storage.BUCKET_NAME+'/'
            if uri in child:
                uri = ''
        else:
            uri = ''

        if path_home in child:
            child = child.replace(path_home,'')

        if path_home in father:
            father = father.replace(path_home,'')

        self.init(uri+child)
        self.connect_post_api()
        self.set_schema()

        if self.config['schema'] != 'files':
            self.schema_class.merge(goal=father)
        else:
            dataset = self.Initializer.storage.dataset
            if self.Initializer.storage.type == 'local':
                father = path_home + '/' + father + '/'
            status = self.status()
            for f in status['keys']:
                self.Initializer.storage.copy_file_global(f,father+f.replace(dataset,''))

        self.init(father)
        self.connect_post_api()
        self.set_schema()
        return True
    
    def branch(self, branch_name='new_branch123/', branch_type='copy', branch_title=''):
        if not self.in_version:
            if branch_name[-1] != '/':
                branch_name += '/'
            if self.config['schema'] != 'files':
                trial = self.schema_class.branch(branch_name=branch_name,type_=branch_type)
                if trial:
                    if self.config['type'] == 's3':
                        uri = 's3://'+self.Initializer.storage.BUCKET_NAME+'/'
                    elif self.config['type'] == 'gcs':
                        uri = 'gs://'+self.Initializer.storage.BUCKET_NAME+'/'
                    else:
                        uri = ''

                    if uri in branch_name:
                        branch_name = branch_name.replace(uri,'')

                    schme = self.config['schema']

                    if branch_name[-1] != '/':
                        branch_name += '/'

                    self.init(uri+branch_name)

                    if branch_title == '':
                        branch_title = branch_name

                    self.reconnect_post_web(branch_title, schme)
                    self.set_schema()
                    self.start_check()
                    self.commit('created branch ' + branch_name)
                    return True
                else:
                    return False
            else:
                type_ = branch_type
                dataset = self.Initializer.storage.dataset
                if self.Initializer.storage.type == 'local':
                    branch_name = path_home + '/' + branch_name + '/'

                status = self.status()
                versions = []
                if len(versions) == 0:
                    for f in status['keys']:
                        versions.append('current')
                idx = 0
                if type_ == 'copy':
                    for f in status['keys']:
                        if versions[idx] == 'current':
                            self.Initializer.storage.copy_file_global(f,branch_name+f.replace(dataset,''))
                        else:
                            self.Initializer.storage.copy_file_global(self.Initializer.prefix_diffs+f+'/'+str(int(versions[idx])).zfill(10),branch_name+f)
                        idx += 1
                else:
                    for f in self.status['keys']:
                        if versions[idx] == 'current':
                            self.Initializer.storage.remove_file_global(dataset+f)
                        else:
                            self.Initializer.storage.copy_file_global(self.Initializer.prefix_diffs+f+'/'+str(int(versions[idx])).zfill(10),branch_name+f)
                        idx += 1

                if self.config['type'] == 's3':
                    uri = 's3://'+self.Initializer.storage.BUCKET_NAME+'/'
                elif self.config['type'] == 'gcs':
                    uri = 'gs://'+self.Initializer.storage.BUCKET_NAME+'/'
                else:
                    uri = ''

                schme = self.config['schema']

                if branch_name[-1] != '/':
                    branch_name += '/'

                self.init(uri+branch_name)

                self.reconnect_post_web(branch_name, schme)
                self.set_schema()
                self.start_check()
                self.commit('created branch ' + branch_name)
        else:
            dataset = self.Initializer.storage.dataset
            if self.Initializer.storage.type == 'local':
                branch_name = path_home + '/' + branch_name + '/'

            status = self.status()
            idx = 0
            for f in status['keys']:
                f_base = os.path.dirname(f.replace(self.Initializer.prefix_diffs,''))
                self.Initializer.storage.copy_file_global(f,branch_name+f_base.replace(dataset,''))
                idx += 1

            if self.config['type'] == 's3':
                uri = 's3://'+self.Initializer.storage.BUCKET_NAME+'/'
            elif self.config['type'] == 'gcs':
                uri = 'gs://'+self.Initializer.storage.BUCKET_NAME+'/'
            else:
                uri = ''

            schme = self.config['schema']

            if branch_name[-1] != '/':
                branch_name += '/'

            self.init(uri+branch_name)

            self.reconnect_post_web(branch_name, schme)
            self.set_schema()
            self.start_check()
            self.commit('created branch ' + branch_name)

            return True

    def schema_metadata(self):
        if self.config['schema'] != 'files':
            return self.schema_class.get_metadata()
        else:
            return {}

    def apply_filter(self):
        if self.config['schema'] != 'files':
            metapath = self.schema_class.meta_path
            return json.load(self.Initializer.storage.load_file_global(metapath))
        else:
            return {}

    def connect_post_web(self, name='My Dataset', keys={}, schema='files'):
        
        print('connecting to '+name)
        self.config['schema'] = schema
        config = self.config

        self.storage_name = config['storage']
        self.dataset_name = config['dataset']
        
        try: 
            datasets = self.get_datasets()
        except:
            datasets = {}

        if config['type'] == 'local':
            cloud = Local()
            cloud.create_dataset(config['dataset'],verbose=True)
            self.Initializer = Initializer(cloud)
        elif config['type'] == 's3':
            cloud = S3Bucket(config['bucket'])
            cloud.connect_bucket_api(keys)
            cloud.create_dataset(config['dataset'])
            self.Initializer = Initializer(cloud)
        elif config['type'] == 'gcs':
            cloud = GCSBucket(config['bucket'])
            cloud.connect_bucket_api(self.key_bin)
            cloud.create_dataset(config['dataset'])
            self.Initializer = Initializer(cloud)
        else:
            self.Initializer = None

        datasets[self.storage_name] = {'storage': self.storage_name, 'name': name, 'type': config['type'], 'schema': config['schema']}
        self.set_datasets(datasets)
        
        self.set_config()
        return True

    def reconnect_post_web(self, name='My Dataset', schema='files'):
        
        print('connecting to '+name)
        self.config['schema'] = schema
        config = self.config

        self.storage_name = config['storage']
        self.dataset_name = config['dataset']
        
        try: 
            datasets = self.get_datasets()
        except:
            datasets = {}

        if config['type'] == 'local':
            cloud = Local()
            cloud.create_dataset(config['dataset'],verbose=True)
            self.Initializer = Initializer(cloud)
        elif config['type'] == 's3':
            cloud = S3Bucket(config['bucket'])
            cloud.reconnect_bucket_api()
            cloud.create_dataset(config['dataset'])
            self.Initializer = Initializer(cloud)
        elif config['type'] == 'gcs':
            cloud = GCSBucket(config['bucket'])
            cloud.reconnect_bucket_api()
            cloud.create_dataset(config['dataset'])
            self.Initializer = Initializer(cloud)
        else:
            self.Initializer = None

        is_not_there = True
        for s in datasets.keys():
            if datasets[s]['storage'] == self.storage_name:
                is_not_there = False

        datasets[self.storage_name] = {'storage': self.storage_name, 'name': name, 'type': config['type'], 'schema': config['schema']}
        self.set_datasets(datasets)
        self.set_config()
        return True

    def connect_post_cli(self):
        config = self.config

        self.storage_name = config['storage']
        self.dataset_name = config['dataset']

        datasets = self.get_datasets()
        try: 
            datasets = self.get_datasets()
        except:
            datasets = {}
        
        try:
            self.config['schema'] = datasets[self.storage_name]['schema']
        except:
            self.config['schema'] = 'files'

        if config['type'] == 'local':
            cloud = Local()
            cloud.create_dataset(config['dataset'], verbose=True)
            self.Initializer = Initializer(cloud)
        elif config['type'] == 's3':
            cloud = S3Bucket(config['bucket'])
            cloud.connect_bucket(verbose=True)
            cloud.create_dataset(config['dataset'])
            self.Initializer = Initializer(cloud)
        elif config['type'] == 'gcs':
            cloud = GCSBucket(config['bucket'])
            cloud.connect_bucket(verbose=True)
            cloud.create_dataset(config['dataset'])
            self.Initializer = Initializer(cloud)
        else:
            self.Initializer = None

        self.set_config()
        return True

    def connect_post_api(self, cli=False):
        config = self.config

        self.storage_name = config['storage']
        self.dataset_name = config['dataset']

        datasets = self.get_datasets()
        try:
            self.config['schema'] = datasets[self.storage_name]['schema']
        except:
            self.config['schema'] = 'files'

        if config['type'] == 'local':
            cloud = Local()
            cloud.create_dataset(config['dataset'], verbose=True)
            self.Initializer = Initializer(cloud)
            self.dataset_name = config['dataset']
            self.storage_name = config['storage']
        elif config['type'] == 's3':
            cloud = S3Bucket(config['bucket'])
            if cli:
                cloud.connect_bucket()
            else:
                cloud.reconnect_bucket_api()
            cloud.create_dataset(config['dataset'])
            self.Initializer = Initializer(cloud)
        elif config['type'] == 'gcs':
            cloud = GCSBucket(config['bucket'])
            if cli:
                cloud.connect_bucket(verbose=True)
            else:
                cloud.reconnect_bucket_api()
            cloud.create_dataset(config['dataset'])
            self.Initializer = Initializer(cloud)
        else:
            self.Initializer = None
        self.set_config()
        return True

    def disconnect_dataset(self, storage=''):
        assert(len(storage) > 0)
        datasets = self.get_datasets()
        print('disconnecting from ' + storage)

        for s in datasets.keys():
            if datasets[s]['storage'] == storage:
                del datasets[s]
                self.set_datasets(datasets)
                return True
        print('run \'stack datasets\' to see other available datasets')
        return False

    def upload_file_binary(self, filename='', binary=''):
        assert(filename != '')
        assert(binary != '')
        add_from_binary(self.Initializer, filename, binary)
        return True

    def upload_file_local_path(self, filename='', target_path=''):
        assert(filename != '')
        add(self.Initializer,[filename],target_path)
        return True

    def start_check(self):
        try:
            return self.Initializer.start_check()
        except:
            return False

    def get_uri(self):
        datasets = self.get_datasets()
        dataset_meta = datasets[self.storage_name]
        return {'storage': self.storage_name, 'schema': dataset_meta['schema'], 'dataset': dataset_meta['name'], 'storage_dataset': self.Initializer.storage.dataset}

    def add(self, path, subpath=''):
        self.reset_version()
        if len(subpath)>1:
            if subpath[-1] != '/':
                subpath = subpath + '/'
        add(self.Initializer,[path],subpath)
        return True

    def pull_all(self,version = 'current'):
        print('downloading files from last commit')
        metapath = self.Initializer.prefix_meta + 'current.json'
        current = json.load(self.Initializer.storage.load_file_global(metapath))
        pull(self.Initializer, current['keys'], version)
        return True

    def pull(self, file, version = 'current'):
        if file == '.' or file == 'all':
            print('downloading all contents')
            return self.pull_all(version)
        else:
            pull(self.Initializer, [file], version)
        return True

    def pull_file(self, file, version = 'current'):
         
        if version == 'current':
            # saves each file
            for key in file:
                binary = self.Initializer.storage.load_file_global(key)
        else:
            gtfo = False
            # finds the commit of interest
            metapath = self.Initializer.prefix_meta+'history.json'
            history = self.Initializer.load(self.Initializer.storage.load_file_global(metapath))
            for key in file:
                if key[-1] == '/':
                    print('Do not pull directories')
                    return False
                for i in range(len(history),int(version)-1,-1):
                    for commit in history[str(i)]['commits']:
                        # reads each file version
                        if self.Initializer.storage.type == 'local':
                            cmit = json.load(self.Initializer.storage.load_file_global(commit))
                        else:
                            cmit = json.load(self.Initializer.storage.load_file_global(commit))
                        if str(cmit['version']) == version and cmit['key'] == key:
                            if cmit['type'] != 'remove':
                                key = self.Initializer.prefix_diffs + key + '/' + str(cmit['version']).zfill(10)
                                binary = self.Initializer.storage.load_file_global(key).read()
                                self.Initializer.storage.reset_buffer()
                            gtfo = True
                        if gtfo:
                            break
                if gtfo:
                    gtfo = False
                    break

        self.Initializer.storage.reset_buffer()
        file_array = {binary: binary, key: file}

        return file_array

    def commits_version(self, version = 2, l = 5, page = 0):
        assert(int(l) > 0)
        assert(int(version) >= 0)
        assert(int(page) >= 0)

        metapath = self.Initializer.prefix_meta+'history.json'
        history = json.load(self.Initializer.storage.load_file_global(metapath))
        self.Initializer.storage.reset_buffer()

        response = {}
        idx = 0

        i_p = int(page)*int(l)
        i_f = min((int(page)+1)*int(l),len(history[str(int(version))]['commits']))

        # goes over the commits
        for i in range(i_p, i_f):
            # reads each file version
            commit = history[str(int(version))]['commits'][i]
            if self.Initializer.storage.type == 'local':
                cmit = json.load(self.Initializer.storage.load_file_global(commit))
            else:
                cmit = json.load(self.Initializer.storage.load_file_global(commit))
            self.Initializer.storage.reset_buffer()
            response[idx] = {'key': cmit['key'], 'source': cmit['source'], 'date': history[str(int(version))]['date'], 'comment': cmit['comment']}
            idx = idx + 1

        return {'commits': response, 'len': len(history[str(int(version))]['commits'])}

    def key_versions(self, key = '', l = 5, page = 0):
        if self.config['schema'] == 'spacy_ner':
            schema = self.schema_class.get_schema()
            key = schema[key]['filename']
        
        if self.Initializer.prefix_diffs in key:
            key = os.path.dirname(key.replace(self.Initializer.prefix_diffs,''))

        assert(int(l) > 0)
        assert(int(page) >= 0)

        key_hist = get_key_history(self.Initializer, key)
        response = {}
        i_p = len(key_hist) - int(page)*int(l)
        i_f = max(len(key_hist) - int(l)*(int(page)+1),0)

        idx = 0

        # goes over the commits
        for i in range(i_p, i_f, -1):
            # reads each file version
            response[idx] = key_hist[str(i)]
            response[idx]['file'] = 'raw'
            idx = idx+1

        return {'commits': response, 'len': len(key_hist)}

    def get_tags(self, key):
        if self.config['schema'] != 'files':
            return self.schema_class.get_tags(key)
        else:
            return []

    def add_tag(self, key, tag):
        if self.config['schema'] != 'files':
            self.schema_class.add_tag(key, tag)
        return True
    
    def add_slice(self, slice_name):
        if self.config['schema'] != 'files':
            self.schema_class.add_slice(slice_name)
        return True

    def remove_slice(self, slice_name):
        if self.config['schema'] != 'files':
            self.schema_class.remove_slice(slice_name)
            self.schema_class.apply_filters()
        return True

    def reset_slices(self):
        if self.config['schema'] != 'files':
            self.schema_class.sliced = False
            self.schema_class.selected_slices = []
            self.schema_class.apply_filters()
        return True
    
    def get_slices(self):
        if self.config['schema'] != 'files':
            return self.schema_class.get_slices()
        return {}
    
    def select_slices(self, slices = []):
        self.reset_version()
        if self.config['schema'] != 'files':
            self.schema_class.select_slice(slices=slices)
        return True

    def selection_add_tag(self, keys, tag):
        if self.config['schema'] != 'files':
            self.schema_class.add_many_tag(keys, tag)
        return True

    def get_readme(self):
        ls, _ = self.Initializer.storage.load_dataset_list()
        matches = [match for match in ls if 'readme' in match.lower()]
        if len(matches) == 0:
            return ''
        else:
            return self.Initializer.storage.load_file_global(matches[0]).read()
    
    def selection_remove_all_tags(self, keys):
        if self.config['schema'] != 'files':
            self.schema_class.many_remove_all_tags(keys)
        return True

    def remove_tag(self, key, tag):
        if self.config['schema'] != 'files':
            self.schema_class.remove_tag(key, tag)
        return True

    def remove_all_tags(self, key):
        if self.config['schema'] != 'files':
            self.schema_class.remove_all_tags(key)
        return True

    def label_versions(self, key, l = 5, page = 0):
        if self.config['schema'] == 'yolo':
            if not self.in_version and self.Initializer.prefix_diffs in key:
                key = os.path.dirname(key.replace(self.Initializer.prefix_diffs,''))
            labels_key = self.schema_class.get_labels_filename(key)
            if self.Initializer.prefix_diffs in labels_key:
                labels_key = os.path.dirname(labels_key.replace(self.Initializer.prefix_diffs,''))
            labels_hist = get_key_history(self.Initializer, labels_key)
            response = {}
            idx = 0
            i_p = len(labels_hist) - int(page)*int(int(l))
            i_f = max(len(labels_hist) - int(int(l))*(int(page)+1),0)

            # goes over the commits
            for i in range(i_p, i_f, -1):
                # reads each file version
                response[idx] = labels_hist[str(i)]
                response[idx]['file'] = 'label'
                idx = idx+1

        elif self.config['schema'] == 'labelbox':
            labels_key = self.schema_class.get_labels_filename()
            if self.in_version:
                labels_key = os.path.dirname(labels_key.replace(self.Initializer.prefix_diffs,''))
            labels_hist = get_key_history(self.Initializer, labels_key)
            response = {}
            idx = 0
            i_p = len(labels_hist) - int(page)*int(int(l))
            i_f = max(len(labels_hist) - int(int(l))*(int(page)+1),0)

            # goes over the commits
            for i in range(i_p, i_f, -1):
                # reads each file version
                response[idx] = labels_hist[str(i)]
                response[idx]['file'] = 'label'
                idx = idx+1
        elif self.config['schema'] == 'spacy_ner':
            schema = self.schema_class.get_schema()
            labels_key = schema[key]['filename']
            labels_hist = get_key_history(self.Initializer, labels_key)
            response = {}
            idx = 0
            i_p = len(labels_hist) - int(page)*int(int(l))
            i_f = max(len(labels_hist) - int(int(l))*(int(page)+1),0)

            # goes over the commits
            for i in range(i_p, i_f, -1):
                # reads each file version
                response[idx] = labels_hist[str(i)]
                response[idx]['file'] = 'label'
                idx = idx+1
        else:
            labels_key = key
            if self.in_version:
                labels_key = os.path.dirname(labels_key.replace(self.Initializer.prefix_diffs,''))
            labels_hist = get_key_history(self.Initializer, key)
            response = {}
            i_p = len(labels_hist) - int(page)*int(l)
            i_f = max(len(labels_hist) - int(l)*(int(page)+1),0)

            idx = 0

            # goes over the commits
            for i in range(i_p, i_f, -1):
                # reads each file version
                response[idx] = labels_hist[str(i)]
                response[idx]['file'] = 'raw'
                idx = idx+1
        return {'commits': response, 'len': len(labels_hist), 'keyId': labels_key}

    def status(self):
        if self.config['schema'] != 'files':
            if self.schema_class.filtered:
                return self.schema_class.status
            else:
                return self.schema_class.read_all_files()
        else:
            if self.filtered:
                if self.in_version:
                    if self.current != None:
                        schema = self.current
                    else:
                        schema = self.Initializer.load_current_version(self.selected_version)
                        self.current = schema
                else:
                    if self.current != None:
                        schema = self.current
                    else:
                        schema = self.Initializer.load_current()
                        self.current = schema
                status = {'keys': [], 'lm': []}

                idx = 0
                for key in schema['keys']:
                    for f in self.filters:
                        for filt in self.filters[f]:
                            if filt == 'name':
                                if self.filters[f]['name'] in key:
                                    status['keys'].append(schema['keys'][idx])
                                    status['lm'].append(schema['lm'][idx])
                    idx += 1
                return status
            else:
                return self.Initializer.load_current()

    def get_next_key(self,key):
        status = self.status()
        idx = status['keys'].index(key)
        if idx == len(status['keys']) - 1:
            return status['keys'][0]
        else:
            return status['keys'][idx + 1]

    def get_prev_key(self,key):
        status = self.status()
        idx = status['keys'].index(key)
        if idx == 0:
            return status['keys'][-1]
        else:
            return status['keys'][idx - 1]

    def remove(self, key, subpath=''):
        if len(subpath)>1:
            if subpath[-1] != '/':
                subpath = subpath + '/'
        key = key.replace(self.Initializer.storage.dataset,'')
        remove(self.Initializer,[key],subpath)
        return True

    def remove_commit(self, version = '-1'):
        self.reset_version()
        assert(int(version) >= 0)
        
        metapath = self.Initializer.prefix_meta+'history.json'
        history = json.load(self.Initializer.storage.load_file_global(metapath))
        self.Initializer.storage.reset_buffer()

        for i in range(len(history[str(int(version))]['commits'])):
            commit = history[str(int(version))]['commits'][i]
            if self.Initializer.storage.type == 'local':
                cmit = json.load(self.Initializer.storage.load_file_global(commit))
            else:
                cmit = json.load(self.Initializer.storage.load_file_global(commit))
            self.Initializer.storage.reset_buffer()
            remove_global(self.Initializer, [cmit['diff']])

        return True

    def remove_key_diff(self, key, version):
        if not self.in_version:
            remove_diff(self.Initializer,key,int(version))
        return True

    
    def set_filters(self, filters):
        if self.config['schema'] != 'files':
            self.schema_class.apply_filters(filters)
        else:
            self.filters = filters
            self.filtered = True
        return True

    def reset_filters(self):
        if self.config['schema'] != 'files':
            self.schema_class.filtered = False
            self.schema_class.apply_filters()
        else:
            self.filtered = False
        return True

    def get_versions(self):
        versions = self.Initializer.load_versions()
        if self.in_version:
            versions['current_v'] = self.selected_version
        else:
            versions['current_v'] = -1
        return versions

    def add_version(self, label = ''):
        add_version(self.Initializer, label)
        if self.config['schema'] != 'files':
            self.schema_class.copy_schema_to_latest_version_checkpoint()
        return True

    def select_version(self, version):
        versions = self.Initializer.load_versions()
        if version in versions.keys():
            self.in_version = True
            self.selected_version = version
            self.version_keys = json.load(self.Initializer.storage.load_file_global(versions[version]['path']))
            if self.config['schema'] != 'files':
                self.schema_class.schema = None
                self.schema_class.select_version(version, versions[version]['schema_path'], self.version_keys)
                self.schema_class.get_schema()
                self.schema_class.read_all_files()
            self.current = self.Initializer.load_current_version(version)
            return True
        else:
            self.reset_version()
            return False

    def reset_version(self):
        if self.in_version:
            self.in_version = False
            self.version_keys = None
            self.current = self.Initializer.load_current()
            self.filtered = False
            if self.config['schema'] != 'files':
                self.schema_class.schema = None
                self.schema_class.in_version = False
                self.schema_class.version_keys = None
                self.schema_class.reset_to_current_version()
                self.schema_class.read_all_files()
                self.filtered = False
        return True

    def get_hierarchy(self):
        metapath = self.Initializer.prefix_meta + 'hierarchy.json'
        try:
            return json.load(self.Initializer.storage.load_file_global(metapath))
        except:
            self.Initializer.storage.add_file_from_binary_global(metapath,io.BytesIO(json.dumps({'parent': '', 'children': []}).encode('ascii')))
            return {'parent': '', 'children': []}

    def set_current_hierarchy(self, hierarchy):
        metapath = self.Initializer.prefix_meta + 'hierarchy.json'
        self.Initializer.storage.add_file_from_binary_global(metapath,io.BytesIO(json.dumps(hierarchy).encode('ascii')))
        self.Initializer.storage.reset_buffer()
        return True
    
    def set_hierarchy(self, parent='', child=''):
        metapath = self.Initializer.prefix_meta + 'hierarchy.json'
        try:
            hierarchy = json.load(self.Initializer.storage.load_file_global(metapath))
        except:
            hierarchy = {'parent': '', 'children': []}

        if self.config['type'] == 's3':
            uri = 's3://'+self.Initializer.storage.BUCKET_NAME+'/'
        elif self.config['type'] == 'gcs':
            uri = 'gs://'+self.Initializer.storage.BUCKET_NAME+'/'
        else:
            uri = ''

        parent = parent.replace(path_home,'')
        child = child.replace(path_home,'')

        if parent != '':
            if not uri in parent:
                parent = uri + parent
            hierarchy['parent'] = parent
        if child != '':
            if not uri in child:
                child = uri + child
            hierarchy['children'].append(child)
        self.Initializer.storage.add_file_from_binary_global(metapath,io.BytesIO(json.dumps(hierarchy).encode('ascii')))
        self.Initializer.storage.reset_buffer()
        return True

    def remove_child(self, child=''):
        metapath = self.Initializer.prefix_meta + 'hierarchy.json'
        try:
            hierarchy = json.load(self.Initializer.storage.load_file_global(metapath))
        except:
            hierarchy = {'parent': '', 'children': []}
        if self.config['type'] == 's3':
            uri = 's3://'+self.Initializer.storage.BUCKET_NAME+'/'
        elif self.config['type'] == 'gcs':
            uri = 'gs://'+self.Initializer.storage.BUCKET_NAME+'/'
        else:
            uri = ''

        if child != '':
            if not uri in child:
                child = uri + child
            if child in hierarchy['children']:
                hierarchy['children'].remove(child)
        self.Initializer.storage.add_file_from_binary_global(metapath,io.BytesIO(json.dumps(hierarchy).encode('ascii')))
        self.Initializer.storage.reset_buffer()
        return True

    def remove_parent(self):
        metapath = self.Initializer.prefix_meta + 'hierarchy.json'
        try:
            hierarchy = json.load(self.Initializer.storage.load_file_global(metapath))
        except:
            hierarchy = {'parent': '', 'children': []}
        hierarchy['parent'] = ''
        self.Initializer.storage.add_file_from_binary_global(metapath,io.BytesIO(json.dumps(hierarchy).encode('ascii')))
        self.Initializer.storage.reset_buffer()
        return True

    def add_child_to_current(self, child=''):
        metapath = self.Initializer.prefix_meta + 'hierarchy.json'
        try:
            hierarchy = json.load(self.Initializer.storage.load_file_global(metapath))
            self.Initializer.storage.reset_buffer()
        except:
            hierarchy = {'parent': '', 'children': []}
        if self.config['type'] == 's3':
            uri = 's3://'+self.Initializer.storage.BUCKET_NAME+'/'
        elif self.config['type'] == 'gcs':
            uri = 'gs://'+self.Initializer.storage.BUCKET_NAME+'/'
        else:
            uri = ''

        if child != '':
            if not uri in child:
                child = uri + child
            if not child in hierarchy['children']:
                hierarchy['children'].append(child)
        self.Initializer.storage.add_file_from_binary_global(metapath,io.BytesIO(json.dumps(hierarchy).encode('ascii')))
        self.Initializer.storage.reset_buffer()
        return True

    def add_parent_to_current(self, parent=''):
        metapath = self.Initializer.prefix_meta + 'hierarchy.json'
        try:
            hierarchy = json.load(self.Initializer.storage.load_file_global(metapath))
        except:
            hierarchy = {'parent': '', 'children': []}
        if self.config['type'] == 's3':
            uri = 's3://'+self.Initializer.storage.BUCKET_NAME+'/'
        elif self.config['type'] == 'gcs':
            uri = 'gs://'+self.Initializer.storage.BUCKET_NAME+'/'
        else:
            uri = ''

        if parent != '':
            if not uri in parent:
                parent = uri + parent
            hierarchy['parent'] = parent
        self.Initializer.storage.add_file_from_binary_global(metapath,io.BytesIO(json.dumps(hierarchy).encode('ascii')))
        self.Initializer.storage.reset_buffer()
        return True

    def remove_key_full(self, key):
        remove_full(self.Initializer,key)
        return True

    def commit(self, comment='',verbose=True):
        self.reset_version()
        res, added, modified, deleted = commit(self.Initializer, comment)
        if verbose:
            if res: 
                print('sync done!')
            else:
                print('already up-to-date')
        if self.config['schema'] != 'files':
            self.schema_class.update_schema_file(added, modified, deleted)

        return True

    def load_commit_metadata(self, commit_file):
        try:
            return json.load(self.Initializer.storage.load_file_global(commit_file))
        except:
            return {}

    def check_if_setup(self):
        self.Initializer.setup_dataset()
        print('setup complete!')
        return True

    def load_file_binary(self, file, version='current'):
        if version=='current':
            print('loading ' + file + '...')
            return self.Initializer.storage.load_file_global(file)
        elif int(version) >= 1:
            print('loading ' + file + ' version '+ version +'...')
            path = self.Initializer.prefix_diffs + file + '/' + str(int(version)).zfill(10)
            return self.Initializer.storage.load_file_global(path)
        else:
            assert(False)

    def load_csv_binary(self, file, row_p, col_p, version='current'):
        assert(int(row_p)>=0)
        assert(int(col_p)>=0)

        N_rows = 10
        N_cols = 10

        if version=='current':
            print('loading ' + file + '...')
            data  = self.Initializer.storage.load_file_global(file)
        elif int(version) >= 1:
            print('loading ' + file + ' version '+ version +'...')
            data  =  self.Initializer.storage.load_file_global(self.Initializer.prefix_diffs+ file + '/' + str(int(version)).zfill(10))
        else:
            assert(False)

        df1 = pd.read_csv(data, encoding='unicode_escape', encoding_errors='backslashreplace', lineterminator='\n')

        total_cols = len(df1.axes[1])
        bot_col = min(N_cols*int(col_p),total_cols)
        top_col = min(N_cols*(int(col_p)+1),total_cols)
        
        total_rows = len(df1.axes[0])
        bot_row = min(N_rows*int(row_p), total_rows)
        top_row = min(N_rows*(int(row_p)+1),total_rows)
        
        df1 = df1.iloc[bot_row:top_row,bot_col:top_col]
        return StringIO(df1.to_csv(index=False, header=False)), total_rows/N_rows, total_cols/N_cols

    def load_csv_diff_metadata(self, file, v1, v2):
        if v1 == 'current':
            d1 = self.Initializer.storage.load_file_global(file)
        else:
            d1 = self.Initializer.storage.load_file_global(self.Initializer.prefix_diffs + file + '/' + str(int(v1)).zfill(10))

        if v2 == 'current':
            d2 = self.Initializer.storage.load_file_global(file)
        else:
            d2 = self.Initializer.storage.load_file_global(self.Initializer.prefix_diffs + file + '/' + str(int(v2)).zfill(10))

        from csv_diff import load_csv, compare
        import codecs

        csv1 = load_csv(codecs.getreader("utf-8")(d1))
        csv2 = load_csv(codecs.getreader("utf-8")(d2))
        diff = compare(csv2, csv1)

        # converts diff to json
        diff_meta = {}
        diff_meta['additions'] = len(diff['added'])
        diff_meta['deletions'] = len(diff['removed'])
        diff_meta['modifications'] = len(diff['changed'])
        diff_meta['new_cols'] = len(diff['columns_added'])

        return diff_meta

    def load_csv_diff(self, file, v1, v2):
        if v1 == 'current':
            d1 = self.Initializer.storage.load_file_global(file)
        else:
            d1 = self.Initializer.storage.load_file_global(self.Initializer.prefix_diffs + file + '/' + str(int(v1)).zfill(10))

        if v2 == 'current':
            d2 = self.Initializer.storage.load_file_global(file)
        else:
            d2 = self.Initializer.storage.load_file_global(self.Initializer.prefix_diffs + file + '/' + str(int(v2)).zfill(10))

        from csv_diff import load_csv, compare
        import codecs

        csv1 = load_csv(codecs.getreader("utf-8")(d1))
        csv2 = load_csv(codecs.getreader("utf-8")(d2))
        diff = compare(csv2, csv1)

        return diff

    def load_file_binary_bytes(self, file, bi, bf):
        print('loading '+file+'...')
        return self.Initializer.storage.load_file_bytes(file,bi,bf)

    def reset(self):
        self.Initializer.remove_setup()
        self.Initializer.setup_dataset()
        print('setup complete!')
        return True

    def revert(self, version=0):
        self.reset_version()
        assert(version != '')
        revert_commit(self.Initializer, int(version))
        self.commit('reverted to version ' + str(version))

    def revert_file(self, key, version):
        try: 
            revert_file(self.Initializer, key, int(version))
            self.Initializer.storage.reset_buffer()
            return True
        except:
            return False

    def history(self):
        metapath = self.Initializer.prefix_meta+'history.json'
        return json.load(self.Initializer.storage.load_file_global(metapath))

    def last_n_commits(self, n = 0):
        metapath = self.Initializer.prefix_meta+'history.json'
        history = json.load(self.Initializer.storage.load_file_global(metapath))
        self.Initializer.storage.reset_buffer()

        response = {}
        idx = 0

        # goes over history
        for i in range(len(history),0,-1):
            for commit in history[str(i)]['commits']:
                # reads each file version
                if self.Initializer.storage.type == 'local':
                    cmit = json.load(self.Initializer.storage.load_file_global(commit))
                else:
                    cmit = json.load(self.Initializer.storage.load_file_global(commit))

                response[idx] = {'source': cmit['source'], 'date': history[str(i)]['date'], 'comment': cmit['comment']}

                idx += 1
                if int(idx) >= int(n):
                    break
            if int(idx) >= int(n):
                    break

        return response

    def diff(self, v1, v0, file=''):
        print_diff(self.Initializer, v1, v0, file)
        return True

    def server_init_experiment(self, uri, project):
        if path_home in uri:
            uri = uri.replace(path_home,'')
        if uri != self.config['storage']:
            self.init(uri)
            self.connect_post_api()
            self.set_schema()

        self.projects = projects(init=self.Initializer, project=project)
        self.projects.init_project()

    def server_add_log(self, data):
        assert(not self.projects is None)
        self.projects.add_log(data)

    def server_add_prediction(self, data):
        assert(not self.projects is None)
        self.projects.add_prediction(data)

    def server_get_models(self):
        assert(not self.projects is None)
        return self.projects.models()

    def server_upload_model(self, model, label):
        assert(not self.projects is None)
        return self.projects.add_model(model, label)

    def server_get_model(self, label):
        assert(not self.projects is None)
        return self.projects.get_model(label)

    def server_logout_experiment(self):
        assert(not self.projects is None)
        self.projects = None

    def get_projects(self):
        prefix_projects = self.Initializer.prefix_meta + 'projects/' + 'projects.json'
        return json.load(self.Initializer.storage.load_file_global(prefix_projects))         

    def get_logs(self, project):
        prefix_projects = self.Initializer.prefix_meta + 'projects/' + project + '/experiments/experiments.json'
        return json.load(self.Initializer.storage.load_file_global(prefix_projects))         
    
    def get_logs_experiment(self, log):
        return json.load(self.Initializer.storage.load_file_global(log))         
    
    def get_predictions_list(self, project):
        prefix_projects = self.Initializer.prefix_meta + 'projects/'
        path = prefix_projects + project + '/predictions.json'
        return json.load(self.Initializer.storage.load_file_global(path))

    def get_prediction(self, prediction):
        return json.load(self.Initializer.storage.load_file_global(prediction))         
        
        



import sys
sys.path.append( '..' )
import stack_api.api_core as api_core
import json
from fastapi import FastAPI, File, UploadFile, Response, Body
from fastapi.middleware.cors import CORSMiddleware
from starlette.responses import StreamingResponse
from pathlib import Path
from pydantic import BaseModel

app = FastAPI()
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"], 
)

class Rate(BaseModel):
    label: str

    @classmethod
    def __get_validators__(cls):
        yield cls.validate_to_json

    @classmethod
    def validate_to_json(cls, value):
        if isinstance(value, str):
            return cls(**json.loads(value))
        return value


from src.comm.docker_ver import *
import os
path_home = os.getenv('LCP_DKR')+'/' if docker_ver() else str(Path.home())
 
# checks if local files are installed
try:
    api = api_core.API()
    initialized = api.start_check()
    api.set_schema()
    # print('doing an initial commit')
    # api.commit('',False)
except:
    print('no config file')
    api = api_core.API()
    print('initializing')
    initilized = api.start_check()
    print('trying')
    try:
        api.set_schema()
    except:
        pass
    print('start check')
    if initilized:
        api.start_check()

# frontend entry-points
@app.post("/init_web/")
async def init_web(data: dict):
    try:
        assert(not '.stack' in data['uri'])
        
        if path_home in data['uri']:
            data['uri'] = data['uri'].replace(path_home,'')

        api.init(data['uri'])
        api.connect_post_web(data['name'], data, data['schema'])
        api.set_schema()
        api.start_check()
        return {'success': True}
    except:
        return {'success': False}

@app.get("/connect/")
async def connect(uri):
    try:
        import time
        t0 = time.time()
        assert(not '.stack' in uri)
        if path_home in uri:
            uri = uri.replace(path_home,'')
        api.init(uri)
        print(f'time to init {time.time() - t0}s')
        t0 = time.time()
        api.connect_post_api()
        print(f'time to connect post api {time.time() - t0}s')
        t0 = time.time()
        api.set_schema()
        print(f'time to set schema {time.time() - t0}s')
        # api.commit()
        t0 = time.time()
        api.reset_version()
        print(f'time to set version {time.time() - t0}s')
        return {'success': True}
    except:
        return {'success': False}

@app.post("/init_gskey/")
def init_gskey(file: UploadFile = File(description="A file read as UploadFile")):
    try:
        api.set_gs_key(file.file)
        return {'success': True}
    except:
        return {'success': False}

@app.post("/set_branch")
def set_branch_api(data: dict):
    try:
        parent = api.Initializer.storage.dataset
        api.set_hierarchy(child = data['branch_name'])
        api.branch(data['branch_name'], data['branch_type'], data['branch_title'])
        api.set_hierarchy(parent = parent)
        return {'success': True}
    except:
        return {'success': False}

@app.get("/get_current_hierarchy")
def get_current_hierarchy_api():
    # try:
    hierarchy = api.get_hierarchy()
    children = []
    for uri in hierarchy['children']:
        children.append({'uri': uri, 'name': api.get_dataset_name(uri)})
    parent = {'uri': hierarchy['parent'], 'name': api.get_dataset_name(hierarchy['parent'])}
    return {'parent': parent, 'children': children}
    # except:
    #     return {'parent': {'uri': '', 'name': ''}, 'children': []}

@app.get("/get_hierarchy")
def get_hierarchy_api(uri):
    try:
        assert(not '.stack' in uri)
        if path_home in uri:
            uri = uri.replace(path_home,'')
        api.init(uri)
        api.connect_post_api()
        api.set_schema()
        return api.get_hierarchy()
    except:
        return {'parent': '', 'children': []}

@app.get("/get_versions")
def get_versions_api():
    try:
        return api.get_versions()
    except:
        return {}

@app.get("/add_version")
def add_version_api(label = ''):
    # try:
    return {'success': api.add_version(label)}
    # except:
    #     return {'success': False}

@app.get("/reset_version")
def reset_version_api():
    try:
        return {'success': api.reset_version()}
    except:
        return {'success': False}

@app.get("/select_version")
def select_version_api(version):
    # try:
    return {'success': api.select_version(str(version))}
    # except:
    #     return {'success': False}

@app.get("/get_branches")
def get_branches_api():
    try:
        hierarchy = api.get_hierarchy()
        res = []
        for uri in hierarchy['children']:
            res.append({'uri': uri, 'name': api.get_dataset_name(uri)})
        return res
    except:
        return []

@app.get("/set_hierarchy")
def set_hierarchy_api(uri_parent, uri_child):
    try:
        assert((not '.stack' in uri_parent) and (not '.stack' in uri_child))

        uri_parent = uri_parent.replace(path_home,'')
        uri_child = uri_child.replace(path_home,'')
        
        api.init(uri_parent)
        api.connect_post_api()
        api.set_schema()
        api.set_hierarchy(child = uri_child)

        api.init(uri_child)
        api.connect_post_api()
        api.set_schema()
        api.set_hierarchy(parent = uri_parent)

        return {'success': True}
    except:
        return {'success': False}

@app.post("/set_current_hierarchy")
def set_current_hierarchy_api(hierarchy: dict):
    try:
        assert('parent' in hierarchy.keys())
        assert('children' in hierarchy.keys())

        assert(type(hierarchy['parent']) is str)
        assert(type(hierarchy['children']) is list)
        
        return {'success': api.set_current_hierarchy(hierarchy)}
    except:
        return {'success': {'parent': '', 'children': []}}

@app.get("/add_child_to_current")
def add_child_to_current_api(child):
    try:
        child = child.replace(path_home,'')
        api.add_child_to_current(child)
        parent = api.Initializer.storage.dataset
        parent = parent.replace(path_home,'')
        try:
            api.init(child)
            api.connect_post_api()
            api.set_schema()
            api.add_parent_to_current(parent)
        except:
            pass
        api.init(parent)
        api.connect_post_api()
        api.set_schema()
        return {'success': True}
    except:
        return {'success': False}

@app.get("/add_parent_to_current")
def add_parent_to_current_api(parent):
    try:
        parent = parent.replace(path_home,'')
        api.add_parent_to_current(parent)
        child = api.Initializer.storage.dataset
        child = child.replace(path_home,'')
        try:
            api.init(parent)
            api.connect_post_api()
            api.set_schema()
            api.add_child_to_current(child)
        except:
            pass
        api.init(child)
        api.connect_post_api()
        api.set_schema()
        return {'success': True}
    except:
        return {'success': False}

@app.get("/current_remove_child")
def current_remove_child(uri=''):
    try:
        uri = uri.replace(path_home,'')
        api.remove_child(child=uri)
        parent = api.Initializer.storage.dataset
        parent = parent.replace(path_home,'')
        try:
            api.init(uri)
            api.connect_post_api()
            api.set_schema()
            api.remove_parent()
        except:
            pass
        api.init(parent)
        api.connect_post_api()
        api.set_schema()
        return {'success': True}
    except:
        return {'success': False}

@app.post("/merge")
def merge_api(data: dict):
    try: 
        api.merge(father=data['father'], child=data['child'])
        api.commit(f"merged branch {data['child']} to {data['father']}")
        return {'success': True}
    except:
        return {'success': False}

@app.get("/merge_child_to_master")
def merge_child_to_master_api(uri):
    try:
        master = api.Initializer.storage.dataset
        api.merge(father=master, child=uri)
        api.commit(f"merged branch {uri} to {master}")
        return {'success': True}
    except:
        return {'success': False}

@app.get("/merge_current_to_master")
def merge_current_to_master_api():
    try: 
        child = api.Initializer.storage.dataset
        hierarchy = api.get_hierarchy()
        if type(hierarchy['parent']) is str:
            if hierarchy['parent'] == '':
                return {'success': False}
            else:
                api.merge(father=hierarchy['parent'], child=child)
                api.commit(f"merged branch {child} to {hierarchy['parent']}")
                return {'success': True}
    except:
        return {'success': False}

@app.get("/disconnect")
def disconnect_api(uri=''):
    try:
        try:         
            api.init(uri)
            api.connect_post_api()
            api.set_schema()
            hierarchy = api.get_hierarchy()
            if len(hierarchy['children']) > 0:
                for child in hierarchy['children']:
                    api.remove_child(child=child)
                    try:
                        api.init(child)
                        api.connect_post_api()
                        api.set_schema()
                        api.remove_parent()
                    except:
                        pass
                api.init(uri)
                api.connect_post_api()
                api.set_schema()
            if type(hierarchy['parent']) is str:
                if hierarchy['parent'] != '':
                    assert(not '.stack' in hierarchy['parent'])
                    if path_home in hierarchy['parent']:
                        hierarchy['parent'] = hierarchy['parent'].replace(path_home,'')
                    try:
                        api.init(hierarchy['parent'])
                        api.connect_post_api()
                        api.set_schema()
                        api.remove_child(child=uri)
                    except: 
                        pass
        except:
            pass    
        return {'success': api.disconnect_dataset(uri)}
    except:
        return {'success': False}

@app.get("/get_datasets")
def get_datasets_api():
    try:
        return api.get_datasets()
    except:
        return {}

@app.get("/uri")
def uri_api():
    try:
        return api.get_uri()
    except:
        return {}

@app.get("/schema")
def schema_api():
    # try:
    return {'value': api.config['schema']}
    # except:
    #     return {'value': 'files'}

@app.get("/history")
def history_api():
    try:
        return api.history()
    except:
        return {}

@app.get('/set_schema')
def set_schema_api():
    api.set_schema()

@app.get('/reset_schema')
def reset_schema_api():
    api.reset_schema()
    return {'success': True}

@app.get('/add_tag')
def add_tag_api(file, tag):
    api.add_tag(file, tag)
    return {'success': True}

@app.get('/remove_tag')
def remove_tag_api(file, tag):
    api.remove_tag(file, tag)
    return {'success': True}

@app.get('/remove_all_tags')
def remove_all_tags_api(file):
    api.remove_all_tags(file)
    return {'success': True}

@app.post('/selection_add_tag')
def selection_add_tag_api(data: list):
    api.selection_add_tag(data[:-1], data[-1])
    return {'success': True}

@app.post('/selection_remove_all_tags')
def selection_remove_all_tags_api(data: list):
    api.selection_remove_all_tags(data)
    return {'success': True}

@app.get('/get_tags')
def get_tags_api(file):
    tags = api.get_tags(file)
    return tags

@app.post("/add_file/")
def add_file_api(file: UploadFile = File(description="A file read as UploadFile")):
    try:
        api.upload_file_binary(file.filename, file.file)
        api.commit('')
        return {'success': True}
    except:
        return {'success': False}

@app.post("/add_multifiles/")
def add_multifiles_api(files: list[UploadFile]):
    try:
        for file in files:
            api.upload_file_binary(file.filename, file.file)
        api.commit('')
        return {'success': True}
    except:
        return {'success': False}

@app.get("/commits_version")
def commits_version_api(version=1,l=5, page=0):
    try:
        return api.commits_version(version, l, page)
    except:
        return {}

@app.get("/key_versions")
def key_versions_api(key='',l=5, page=0):
    # try:
    return api.key_versions(key, l, page)
    # except:
    #     return {}

@app.get("/label_versions")
def label_versions_api(key='',l=5, page=0):
    try:
        return api.label_versions(key, l, page)
    except:
        return {}

@app.get("/last_n_commits")
def last_n_commits_api(n=5):
    try:
        return api.last_n_commits(n)
    except:
        return {}

@app.get("/last_commits_from_hist_api")
def last_commits_from_hist_api(n=1):
    try:
        return api.getHistoryCommits(n)
    except:
        return {}

@app.get("/status")
def status_api():
    # try:
    return api.status()
    # except:
    #     return {}

@app.get("/schema_metadata")
def schema_metadata_api():
    return api.schema_metadata()

@app.get("/current")
def current_api(page=0,max_pp=12):
    # try:
    full_json = api.status()
    idx_i = int(page)*int(max_pp)
    idx_f = (int(page)+1)*int(max_pp)

    current = {'keys': [], 'lm': [], 'len': len(full_json['keys'])}

    current['keys'] = full_json['keys'][idx_i:idx_f]
    current['lm'] = full_json['lm'][idx_i:idx_f]

    return current
    # except:
    #     return {'keys': {}, 'lm': {}}

@app.post("/set_filter/")
def set_filter_api(data: dict):
    return {'success': api.set_filters(data)}

@app.get("/reset_filters")
def reset_filter_api():
    return {'success': api.reset_filters()}

@app.get("/commit_req")
def commit_api(comment=''):
    return {'success': api.commit(comment)}

@app.get("/get_commit_metadata")
def get_commit_meta_api(commit):
    try:
        return api.load_commit_metadata(commit)
    except:
        return {}

@app.get("/pull_file_api")
def pull_file_api(file, version='current'):
    try:
        return StreamingResponse(api.load_file_binary(file, version), media_type="image/png")
    except:
        return Response(content='')

@app.get("/set_bounding_boxes")
def set_bounding_boxes_api(val):
    try:
        return {'success': api.set_bounding_boxes(val)} 
    except:
        return {'success': False}

@app.get("/get_thumbnail")
def get_thumbnail_api(file):
    try:
        import time
        t0 = time.time()
        thumb = api.load_thumbnail(file)
        print(f'time to thumbnail {time.time() - t0}s')
        return StreamingResponse(thumb, media_type="image/png")
    except:
        return Response(content='')

@app.get("/pull_csv_api")
def pull_csv_api(file, row_p=0, col_p=0, version='current'):
    try:
        data, _, _ = api.load_csv_binary(file, row_p, col_p, version)
        return StreamingResponse(data, media_type="image/png")
    except:
        return Response(content='')

@app.get("/pull_csv_metadata_api")
def pull_csv_metadata_api(file, version='current'):
    try:
        _, nr, nc = api.load_csv_binary(file, 0, 0, version)
        return {'rows': nr-1, 'cols': nc-1}
    except:
        return {'rows': 0, 'cols': 0}

@app.get("/remove_key")
def remove_key_api(key):
    try:
        api.remove(key)
        api.commit('')
        return {'sucess': True}
    except:
        return {'sucess': False}

@app.get("/remove_commit")
def remove_commit(version):
    try:
        api.remove_commit(version)
        api.commit('')
        return {'sucess': True}
    except:
        return {'sucess': False}

@app.get("/full_remove_key")
def full_remove_key_api(key):
    try:
        api.remove_full(key)
        api.commit('')
        return {'sucess': True}
    except:
        return {'sucess': False}

@app.get("/remove_key_version")
def remove_key_diff_api(key, version=-1):
    try:
        api.remove_key_diff(key, version)
        api.commit('')
        return {'sucess': True}
    except:
        return {'sucess': False}

@app.get("/get_csv_diff_metadata")
def get_csv_diff_metadata(key, v1='current', v2='current'):
    try:
        return api.load_csv_diff_metadata(key, v1, v2)
    except:
        return {}

@app.get("/get_csv_diff")
def get_csv_diff(key, v1='current', v2='current'):
    try:
        return api.load_csv_diff(key, v1, v2)
    except:
        return {}

@app.get("/revert_key_version")
def revert_key_version_api(key, version=-1, label='raw'):
    # try:
    api.revert_file(key, version)
    api.commit('reverted file ' + key)
    return {'sucess': True}
    # except:
    #     return {'sucess': False}

@app.get("/revert")
async def revert_api(version=0):
    # try:
    api.revert(version)
    api.commit('reverted dataset to version ' + version)
    return {'success': True}
    # except:
    #     return {'success': False}

@app.get("/get_prev_key")
def get_prev_key_api(key):
    return {'key': api.get_prev_key(key)}

@app.get("/get_next_key")
def get_next_key_api(key):
    return {'key': api.get_next_key(key)}

@app.get("/get_labels")
def get_labels_api(filename, version='current'):
    # try:
    return api.get_labels(filename, version)
    # except:
    #     return {}

@app.post("/set_labels")
def set_labels_api(data: dict):
    # try:
    return api.set_labels(data)
    # except:
    #     return {}

@app.get("/add_slice")
def add_slice(slice_name):
    return {'success': api.add_slice(slice_name=slice_name)}

@app.get("/remove_slice")
def remove_slice(slice_name):
    return {'success': api.remove_slice(slice_name=slice_name)}

@app.get("/reset_slices")
def reset_slices():
    return {'success': api.reset_slices()}

@app.post("/select_slice")
def select_slice(slices: list):
    return {'success': api.select_slices(slices=slices)}

@app.get("/get_slices")
def get_slices():
    return api.get_slices()

@app.get("/get_readme")
def get_readme():
    try:
        return Response(content=api.get_readme())
    except:
        return Response(content='')

# stacklib entry-points

@app.get("/init_experiment")
def init_experiment(data: dict):
    api.server_init_experiment(uri=data['uri'],project=data['project'])
    return {'success': True}

@app.post("/add_log")
def add_log(data: dict):
    api.server_add_log(data)
    return {'success': True}

@app.post("/get_models")
def get_models():
    return api.server_get_models()

@app.post("/add_model")
def add_model(data: Rate = Body(...), file: UploadFile = File(...)):
    api.server_upload_model(model=file.file, label=data.label)
    return {'success': True}

@app.get("/add_prediction")
def add_prediction(data: dict):
    return StreamingResponse(api.server_add_prediction(data))

@app.get("/get_model")
def get_model(data: dict):
    return StreamingResponse(api.server_get_model(data['label']))

@app.get("/logout_experiment")
def logout_experiment():
    api.server_logout_experiment()
    return {'success': True}

@app.get("/get_projects")
def get_projects():
    try:
        return api.get_projects()
    except:
        return []

@app.get("/get_logs_list")
def get_logs_list(project):
    return api.get_logs(project)

@app.get("/get_logs_experiment")
def get_logs_experiment(log):
    return api.get_logs_experiment(log)

@app.get("/get_predictions_list")
def get_predictions_list(prediction):
    return api.get_predictions_list(prediction)

@app.get("/get_prediction")
def get_prediction(prediction):
    return api.get_prediction(prediction)


import sys
sys.path.append( '../..' )
import webbrowser
from nylas import APIClient
from dotenv import *

config = dotenv_values()

class Communications(object):
	"""docstring for Communications"""
	def __init__(self, user):
		super(Communications, self).__init__()
		self.email = user.email
		self.nylas = APIClient(
			config['NYLAS_KEY_1'],
			config['NYLAS_KEY_2'],
			config['NYLAS_KEY_3'])

	def requestSupport(self):
		webbrowser.open('https://calendly.com/eigen-team/eigen-team')
		return True

	def sendFeedback(self):
		draft = self.nylas.drafts.create()
		draft.subject = "feedback from user: "+self.email
		message = input('Type your feedback: ')
		draft.body = "Hello, I'm user "+self.email+"\n"+message+"\nBest!"
		draft.to = [{'name': 'B', 'email': 'getstack@mit.edu'}]

		try:
			draft.send()
		except:
			print('failure to send email')
			return False

		return True

	def sendUpdate(self, update):
		draft = self.nylas.drafts.create()
		draft.subject = "Update "+self.update.title
		draft.body = update.body
		draft.to = [{'name': 'B', 'email': self.email}]

		try:
			draft.send()
		except:
			print('failure to send email update')
			return False

		return True

if __name__ == '__main__':
	import sys
	sys.path.append( '../..' )
	from src.user.user import User
	usr = User()
	usr.loginUser()
	comm = Communications(usr)
	comm.sendFeedback()



import socket
import string
import json
import io
from datetime import datetime
import time
from pathlib import Path
import os
from src.comm.docker_ver import *
path_home = os.getenv('LCP_DKR')+'/' if docker_ver() else str(Path.home())

class Initializer(object):
	"""docstring for Initializer"""
	def __init__(self, storage ,user=None):
		super(Initializer, self).__init__()
		self.storage = storage
		self.dataset = storage.load_dataset()
		self.user = user
		self.schema = 'Files'
		
		# prefixes
		if self.storage.type == 'local':
			self.prefix_meta = path_home + '/.stack/' +  self.storage.dataset
			self.prefix_commit = path_home + '/.stack/' + self.storage.dataset + 'commits/'
			self.prefix_history = path_home + '/.stack/' + self.storage.dataset + 'history/'
			self.prefix_diffs = path_home + '/.stack/' + self.storage.dataset + 'diffs/'
			self.prefix_versions = path_home + '/.stack/' + self.storage.dataset + 'versions/'
		else:
			self.prefix_meta = '.stack/'+self.storage.dataset
			self.prefix_commit = '.stack/'+self.storage.dataset+'commits/'
			self.prefix_history = '.stack/'+self.storage.dataset+'history/'
			self.prefix_diffs = '.stack/'+self.storage.dataset+'diffs/'
			self.prefix_versions = '/.stack/' + self.storage.dataset + 'versions/'

		self.start_check()

	def remove_setup(self):
		if self.storage.type == 'local':
			self.storage.remove_file_global(path_home+'/.stack/')
		else:
			self.storage.remove_file_global('.stack/')
		return True

	def load_current(self):
		return json.load(self.storage.load_file_global(self.prefix_meta+'current.json'))

	def load_current_version(self, version):
		versions = self.load_versions()
		keys = json.load(self.storage.load_file_global(versions[str(version)]['path']))
		
		return {'keys': list(keys.values()), 'lm': [versions[str(version)]['date']] * len(list(keys.values()))}
	
	def load_versions(self):
		metapath = self.prefix_versions + 'versions.json'
		try:
			versions = json.load(self.storage.load_file_global(metapath))
		except:
			self.setup_versions()
			versions = json.load(self.storage.load_file_global(metapath))
		return versions

	def remove_dataset_setup(self):
		self.storage.remove_file_global(self.prefix_meta)
		return True

	def start_check(self):
		# checks if the dataset exists
		if not self.verify_setup():
			self.setup_dataset()

		if self.storage.check_if_empty(self.prefix_versions):
			self.setup_versions()

		return True

	def verify_setup(self):
		# verifies if a setup .stack/ directory exists
		if self.storage.check_if_empty(self.prefix_meta):
			return False

		if self.storage.check_if_empty(self.prefix_diffs) and (len(self.dataset) > 0):
			return False

		if self.storage.check_if_empty(self.prefix_commit) and (len(self.dataset) > 0):
			return False

		return True

	def setup_dataset(self):
		# performs all key operations
		self.copy_current()
		self.setup_hierarchy()
		self.setup_diffs()
		self.setup_commits()
		self.setup_history()
		
		return True

	def setup_hierarchy(self):
		metapath = self.prefix_meta + 'hierarchy.json'
		hierarchy = {'parent': '', 'children': []}
		self.storage.add_file_from_binary_global(metapath,io.BytesIO(json.dumps(hierarchy).encode('ascii')))
		self.storage.reset_buffer()
		return True

	def setup_versions(self):
		keys = {}
		for file in self.dataset:
			ver = 1
			keys[file['key']] = self.prefix_diffs + file['key'] + '/' + str(ver).zfill(10)

		ver_path = self.prefix_versions + '/0.json'
		self.storage.add_file_from_binary_global(ver_path,io.BytesIO(json.dumps(keys).encode('ascii')))
		
		time = datetime.now().strftime("%m/%d/%Y %H:%M:%S")
		history = {0: {'path': ver_path, 'schema_path': '', 'date': time, 'label': 'Version 0'}}
		metapath = self.prefix_versions + 'versions.json'
		self.storage.add_file_from_binary_global(metapath,io.BytesIO(json.dumps(history).encode('ascii')))

	def copy_current(self):
		# backup of the current version to keep track of changes
		metapath = self.prefix_meta + 'current.json'
		keys, lm = self.storage.load_dataset_list()
		current = {'keys': keys, 'lm': lm}
		self.storage.add_file_from_binary_global(metapath,io.BytesIO(json.dumps(current).encode('ascii')))
		self.storage.reset_buffer()
		return True

	def setup_diffs(self):
		# adds all files, creates a diff for each
		# located in .stack/meta/diffs/abcde/DIFF_ID

		# stores the diffs of the first commit
		for file in self.dataset:
			# TODO: let's name the diffs with just numbering for now
			# and I pre-append 10 zeros bcs I do not remember
			diff = self.prefix_diffs + file['key'] + '/' + str(1).zfill(10)
			self.storage.copy_file_global(file['key'],diff)		
		return True

	def setup_history(self):
		# meta-data
		# located in .stack/meta/history.json
		""" 
			json = {
				0 : ['c1.json','c2.json'],
				1 : ['c3.json'],
			}
		"""
		commits = []
		for file in self.dataset:
			prefix_commit = self.prefix_commit
			commitpath = prefix_commit + file['key'] + '/' + str(1).zfill(10)
			commits.append(commitpath)

		time = datetime.now().strftime("%m/%d/%Y %H:%M:%S")
		history = {1: {'commits': commits, 'date': time}}
		metapath = self.prefix_meta+'history.json'
		self.storage.add_file_from_binary_global(metapath,io.BytesIO(json.dumps(history).encode('ascii')))

		return True

	def setup_commits(self):
		# meta-data for each commit
		# located in .stack/meta/commits/abcde/COMMIT_ID
		""" 
			json = {
				'file' : 'img1_jpg',
				'id:   : 'abcdefg'
				'diff' : 'dff1',
				'type' : 'add',
				'date' : [...],
				'source' : '0.0.0.0',
				'comment' : '',
			}
		"""

		# checks who is the user
		idx = 0
		if self.user == None:
			username = socket.gethostbyname(socket.gethostname())
		else:
			username = self.user.username

		# stores the medadata of the first commit
		for file in self.dataset:
			# gets the address of the diff
			# TODO: let's name the diffs with just numbering for now
			diff = self.prefix_diffs + file['key'] + '/' + str(1).zfill(10)

			commit = {
				'key'	: file['key'],
				'diff'	: diff,
				'type'	: 'add',
				'version' : 1,
				'date'	: file['last_modified'],
				'source': username,
				'comment' : 'added '+ file['key'],
			}

			history = {}
			history[1] = commit
			commitpath = self.prefix_commit + file['key'] + '/' + str(1).zfill(10)
			self.storage.add_file_from_binary_global(commitpath,io.BytesIO(json.dumps(commit).encode('ascii')))
			
			self.storage.reset_buffer()
			
			histpath = self.prefix_history + file['key'] + '/history.json'
			self.storage.add_file_from_binary_global(histpath,io.BytesIO(json.dumps(history).encode('ascii')))
			self.storage.reset_buffer()

		self.storage.reset_buffer()
		
		return True

	def get_latest_diff_number(self, key):
		# checks all the diffs
		histpath = self.prefix_history + key + '/history.json'
		try:
			history = json.load(self.storage.load_file_global(histpath))
			return len(history.keys())
		except:
			if not self.verify_setup:
				self.setup_dataset()
			return 0

	def get_latest_version_number(self):
		# checks all the diffs
		metapath = self.prefix_versions + 'versions.json'
		try:
			versions = json.load(self.storage.load_file_global(metapath))
			return len(versions.keys())
		except:
			return 0
			
	def get_list_to_compare(self):
		# get new files
		new_files, new_lm = self.storage.load_dataset_list()
		metapath = self.prefix_meta + 'current.json'
		current = json.load(self.storage.load_file_global(metapath))

		return new_files, new_lm, current['keys'], current['lm']

def main():
	import sys
	sys.path.append( '../..' )

	from src.storage.classes.s3 import S3Bucket
	
	cloud = S3Bucket('stacktest123')
	cloud.connect_bucket()
	cloud.create_dataset('dataset1/')
	print('dataset created')

	init = Initializer(cloud)
	print('init created')

	init.setup_diffs()
	print('first diffs added')

	init.setup_commits()
	print('first commits added')

	init.setup_history()
	print('first change added')

	init.storage.copy_file('image.png','image2.png')
	
	nf,_,of,_ = init.get_list_to_compare()
	print('list of new files')
	print(nf)

	print('list of old files')
	print(of)

if __name__ == '__main__':
	main()



# from src.storage import Storage
import os
import logging
from google.cloud import storage
import maskpass
import io
from pathlib import Path
import sys
sys.path.append( '../../../' )
from src.comm.docker_ver import *
path_home = os.getenv('LCP_DKR')+'/' if docker_ver() else str(Path.home())

class GCSBucket(object):
	"""docstring for Storage"""
	def __init__(self, BUCKET_NAME):
		# self = Cloud().__init__()
		self.type = "gcs"
		self.dataset = ""
		self.BUCKET_NAME = BUCKET_NAME
		self.credentials = {}
		self.client = None
		self.bucket = None

	def connect_bucket(self, verbose=False):
		# creates a client 
		key_path = path_home+'/.gs/'+self.BUCKET_NAME+'/gs_key'
		os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = key_path
		try:		
			self.client = storage.Client()
		except:
			# in case the keys are not in the computer
			print('1) Create a service account here: https://console.cloud.google.com/projectselector/iam-admin/serviceaccounts/create')
			print('2) Give it access to Cloud Storage')
			print('3) Go to your new service account/keys')
			print('4) Create a new key in .json format and save it as ~/.gs_key')
			return False

		# checks if the bucket exists
		if verbose:
			print('Connecting to your bucket...')
		buckets = self.client.list_buckets()
		found = False
		for bucket in buckets:
			if bucket.name == self.BUCKET_NAME:
				found = True
				break

		# connects to bucket or finds another one
		if found:
			self.bucket = self.client.get_bucket(self.BUCKET_NAME)
		else:			
			if verbose:
				print('could not find your desired bucket')
				print('connection failed')
				raise Exception('Could not find your desired bucket')
		return True

	def connect_bucket_api(self,binary):
		# reads the gs_key
		print('creating key file')
		key_path = path_home+'/.gs/'+self.BUCKET_NAME+'/gs_key'

		if not os.path.isfile(key_path):
			try:
				os.mkdir(path_home+'/.gs/')
			except:
				pass
			try: 
				os.mkdir(path_home+'/.gs/'+self.BUCKET_NAME+'/')
			except:
				pass

		binary_file = open(key_path, "wb")
		binary_file.write(binary.read())
		binary_file.close()
		
		# sets env variable for google cloud
		print('exporting key')	
		os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = key_path

		print('connecting to client')
		try:
			self.client = storage.Client()
		except:
			return False

		# checks if the bucket exists
		print('Connecting to your bucket...')
		buckets = self.client.list_buckets()
		for bucket in buckets:
			if bucket.name == self.BUCKET_NAME:
				self.bucket = self.client.get_bucket(self.BUCKET_NAME)
				return True
		return False

	def reconnect_bucket_api(self):
		try:
			key_path = path_home+'/.gs/'+self.BUCKET_NAME+'/gs_key'	
			os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = key_path

			print('connecting to client')
			self.client = storage.Client()
		except:
			return False

		# checks if the bucket exists
		print('Connecting to your bucket...')
		buckets = self.client.list_buckets()
		for bucket in buckets:
			if bucket.name == self.BUCKET_NAME:
				self.bucket = self.client.get_bucket(self.BUCKET_NAME)
				return True
		return False

	def create_dataset(self,location):
		# assigns location
		if location[-1] != '/':
			location = location + '/'
		self.dataset = location
		self.raw_location = location
		return True

	def add_file(self,filepath,target_name='',subpath=''):
		# adding new file
		if target_name == '':
			target_name = os.path.basename(filepath)
		if filepath[-1] == '/':
			for root,dirs,files in os.walk(filepath):
				for file in files:
					blob = self.bucket.blob(self.dataset+subpath+os.path.basename(filepath))
					blob.upload_from_filename(os.path.join(root,file))	
			self.reset_buffer()
		else:
			blob = self.bucket.blob(self.dataset+subpath+os.path.basename(filepath))
			blob.upload_from_filename(filepath)
		return True

	def add_file_from_binary(self,filename,f):
		return self.add_file_from_binary_global(self.dataset+filename,f)

	def add_file_from_binary_global(self,filename,f):
		blob = self.bucket.blob(filename)
		blob.upload_from_file(file_obj=f, rewind=True)
		return True

	def remove_file(self,filename):
		return self.remove_file_global(self.dataset+filename)

	def remove_file_global(self,filename):
		# deletes files in path
		if filename[-1] == '/':
			for blob in self.client.list_blobs(self.BUCKET_NAME, prefix=filename, delimiter=None):
				blob.delete()
			self.reset_buffer()
		else:
			blob = self.bucket.blob(filename)
			blob.delete()
		return True

	def load_file(self,filename):
		return self.load_file_global(self.dataset+filename)

	def load_file_bytes(self,filename,bi,bf):
		return self.load_file_global_bytes(self.dataset+filename,bi,bf)

	def load_file_global(self,filename):
		blob = self.bucket.blob(filename)
		file_obj = blob.download_as_string()
		return io.BytesIO(file_obj)

	def load_file_global_bytes(self,filename,bi,bf):
		blob = self.bucket.blob(filename)
		file_obj = blob.download_as_string(start=bi,end=bf)
		return io.BytesIO(file_obj)	

	def load_file_metadata(self,filename,debug=False):
		blob = self.bucket.get_blob(self.dataset + filename)
		if blob.name[-1] != '/':
			metadata = {
				'key' : self.dataset + filename,
				'date_loaded' : blob.time_created.strftime("%m/%d/%Y, %H:%M:%S"),
				'date_added' : blob.time_created.strftime("%m/%d/%Y, %H:%M:%S"),
				'last_modified' : blob.updated.strftime("%m/%d/%Y, %H:%M:%S"),
				'HostId' : blob.id,
			}
			return metadata

	def load_file_metadata_global(self,filename,debug=False):
		blob = self.bucket.get_blob(filename)
		if blob.name[-1] != '/':
			metadata = {
				'key' : self.dataset + filename,
				'date_loaded' : blob.time_created.strftime("%m/%d/%Y, %H:%M:%S"),
				'date_added' : blob.time_created.strftime("%m/%d/%Y, %H:%M:%S"),
				'last_modified' : blob.updated.strftime("%m/%d/%Y, %H:%M:%S"),
				'HostId' : blob.id,
			}
			return metadata

	def list_files_in_path(self,path):
		blobs = self.client.list_blobs(self.BUCKET_NAME, prefix=path, delimiter=None)
		for blob in blobs:
			print(f'-- {blob.name}')
		return True

	def load_dataset(self):
		files = []
		blobs = self.client.list_blobs(self.BUCKET_NAME, prefix=self.dataset, delimiter=None)
		# labels each file
		for blob in blobs:
			if blob.name[-1] != '/':
				metadata = {
					'key' : blob.name,
					'last_modified' : blob.updated.strftime("%m/%d/%Y, %H:%M:%S"),
				}
				files.append(metadata)
		return files

	def load_dataset_list(self):
		# loads all the metadata associated with the dataset
		return self.load_list_in_path(self.dataset)

	def load_list_in_path(self,path):
		# loads all the metadata associated within a path
		keys = []
		last_m = []

		blobs = self.client.list_blobs(self.BUCKET_NAME, prefix=path, delimiter=None)

		# loads each file
		for blob in blobs:
			if blob.name[-1] != '/':
				keys.append(blob.name)
				last_m.append(blob.updated.strftime("%m/%d/%Y, %H:%M:%S"))
			
		return keys, last_m

	def check_if_empty(self,path):
		# checks the number of files in a directory
		blobs = self.client.list_blobs(self.BUCKET_NAME, prefix=path, delimiter=None)
		idx = 0
		for b in blobs:
			idx = idx + 1
		return (idx == 0)

	def load_files_in_dataset(self):
		# lists all the files in the main path of the dataset
		return self.list_files_in_path(self.dataset)

	def copy_file(self,filepath,full_target_name):
		# adding new file
		return self.copy_file_global(self.dataset+filepath,self.dataset+full_target_name)

	def copy_file_global(self,filepath,full_target_name):
		source_blob = self.bucket.blob(filepath)
		blob_copy = self.bucket.copy_blob(source_blob, self.bucket, full_target_name)
		return True

	def get_size_of_file(self, filepath):
		return self.get_size_of_file_global(self.dataset + filepath)

	def get_size_of_file_global(self, filepath):
		return self.bucket.get_blob(filepath).size

	def reset_buffer(self):
		return True

def main():
	# connects a bucket
	bucket = 'gs://stack123'
	cloud = GCSBucket('stack123')
	cloud.connect_bucket()

	# connects to a dataset in the bucket
	cloud.create_dataset('dataset1/')

	# list the files in the bucket
	cloud.list_files_in_path('')
	cloud.load_files_in_dataset()

	# addes on file
	cloud.add_file('../../tests/image.png','image.png')
	cloud.add_file('../../tests/image.png','base/image.png',subpath='base/')
	cloud.reset_buffer()

	# opens one file
	body = cloud.load_file('image.png')
	print(body)

	meta = cloud.load_file_metadata('image.png')
	print(meta)

	# adds two file from binaries
	cloud.add_file_from_binary('extra/image2.png',open("../../tests/image.png", "rb"))
	cloud.add_file_from_binary('extra/image3.png',cloud.load_file('base/image.png'))
	cloud.reset_buffer()

	# removes a file
	cloud.remove_file('extra/image2.png')

	# loads the dataset in memory
	dataset = cloud.load_dataset()
	print(cloud.load_dataset())

	cloud.load_files_in_dataset()
	
	cloud.copy_file_global(cloud.dataset+'extra/image3.png',cloud.dataset+'extra/image4.png')
	cloud.list_files_in_path('')
	print('success')

if __name__ == '__main__':
	main()


import sys
sys.path.append( '../..' )
from src.core.init import Initializer
import socket
import string
import json
import io
from datetime import datetime
import time
import difflib
import re
import os
from pathlib import Path
from src.comm.docker_ver import *
path_home = os.getenv('LCP_DKR')+'/' if docker_ver() else str(Path.home())

def commit(init, comment = ''):
	# get new files
	# print('reading the data list')
	# TODO: Oprimize
	t0 = time.time()
	new_files, new_lm, old_files, old_lm = init.get_list_to_compare()
	# print(f'time get data list {time.time() - t0}s for {len(old_files)} datapoints')

	if comment != '':
		comment = ', ' + comment

	init.storage.reset_buffer()
	# checks who is the user
	idx = 0
	if init.user == None:
		try:
			username = socket.gethostbyname(socket.gethostname())
		except:
			username = 'unknown'
	else:
		username = init.user.username

	commits = []
	toadd = []
	toremove = []

	added = []
	modified = []
	removed = []
	# print('starting a commit')
	t0 = time.time()

	# new_files_s = set(new_files)

	for idx, f in enumerate(old_files):
		# relative file location
		f_relative = f

		# checks if the file was modified or remove
		try:
			idx_ = new_files.index(f_relative)
		except:
			idx_ = -1

		if idx_ > -1:
			if new_lm[idx_] != old_lm[idx]:
				# checks the latest diff and sets the new path
				n = init.get_latest_diff_number(f_relative) + 1
				diff = init.prefix_diffs + f_relative + '/' + str(n).zfill(10)

				# computes the diff and stores it
				if init.storage.type == 'local':
					store_diff(init.storage,diff,f_relative,f)
				else:
					store_diff(init.storage,diff,f_relative,f)

				commit = {
					'key'	: f_relative,
					'diff'	: diff,
					'version' : n,
					'type'	: 'modified',
					'date'	: new_lm[idx_],
					'source': username,
					'comment' : 'modified '+ f_relative + comment,
				}

				print('-- modified '+ f_relative)

				toadd.append(f_relative)
				modified.append(f_relative)
	
				commitpath = init.prefix_commit + f_relative + '/' + str(n).zfill(10)
				commits.append(commitpath)
				init.storage.add_file_from_binary_global(commitpath,io.BytesIO(json.dumps(commit).encode('ascii')))
				update_file_history(init,f_relative,commit)
		else:
			# checks the latest diff and sets the new path
			n = init.get_latest_diff_number(f_relative) + 1
			diff = init.prefix_diffs + f_relative + '/' + str(n).zfill(10)
			init.storage.add_file_from_binary_global(diff,io.BytesIO("".encode('ascii')))

			commit = {
				'key'	: f_relative,
				'diff'	: diff,
				'version' : n,
				'type'	: 'remove',
				'date'	: old_lm[idx],
				'source': username,
				'comment' : 'removed '+ f_relative+comment,
			}

			print('-- removed '+ f_relative)

			toremove.append(f_relative)
			removed.append(f_relative)

			prefix_commit = init.prefix_commit

			commitpath = prefix_commit + f_relative + '/' + str(n).zfill(10)
			commits.append(commitpath)
			init.storage.add_file_from_binary_global(commitpath,io.BytesIO(json.dumps(commit).encode('ascii')))
			update_file_history(init,f_relative,commit)

	# print(f'time to do first comparisons {time.time() - t0}s for {len(old_files)} datapoints')
	init.storage.reset_buffer()

	old_rel_files_s = set(old_files)

	for idx, f in enumerate(new_files):
		# checks if the file was modified or remove
		if not (f in old_rel_files_s):
			if not '.DS_Store' in f:
				# checks the latest diff and sets the new path
				n = init.get_latest_diff_number(f) + 1
				diff = init.prefix_diffs + f + '/' + str(n).zfill(10)

				if init.storage.type == 'local':
					store_diff(init.storage,diff,f,f)
				else:
					store_diff(init.storage,diff,f,f)

				commit = {
					'key'	: f,
					'diff'	: diff,
					'version' : n,
					'type'	: 'add',
					'date'	: new_lm[idx],
					'source': username,
					'comment' : 'added '+ f+comment,
				}

				print('-- added '+ f)			

				toadd.append(f)
				added.append(f)

				prefix_commit = init.prefix_commit
				commitpath = prefix_commit + f + '/' + str(n).zfill(10)
				commits.append(commitpath)
				init.storage.add_file_from_binary_global(commitpath,io.BytesIO(json.dumps(commit).encode('ascii')))
				update_file_history(init,f,commit)
	init.storage.reset_buffer()	

	# print(f'time to do all comparisons {time.time() - t0}s for {len(old_files)} datapoints')

	# updates the current version in .stack
	if len(commits):
		init.copy_current()
		update_history(init,commits)
	init.storage.reset_buffer()

	return (len(commits) > 0), added, modified, removed

def add_version(init, label=''):
		keys = {}
		dataset = init.storage.load_dataset()
		metapath = init.prefix_versions + 'versions.json'
		history = json.load(init.storage.load_file_global(metapath))
		ver_path = init.prefix_versions + f'/{len(history.keys())}.json'
		
		for file in dataset:
			ver = init.get_latest_diff_number(file['key'])
			keys[file['key']] = init.prefix_diffs + file['key'] + '/' + str(ver).zfill(10)

		init.storage.add_file_from_binary_global(ver_path,io.BytesIO(json.dumps(keys).encode('ascii')))
		time = datetime.now().strftime("%m/%d/%Y %H:%M:%S")
		
		if label == '':
			label = f'Checkpoint {len(history.keys())}'

		history[str(len(history.keys()))] = {'path': ver_path, 'schema_path': '', 'date': time, 'label': label}
		init.storage.add_file_from_binary_global(metapath,io.BytesIO(json.dumps(history).encode('ascii')))

def compute_diff(bin1,bin2):
	"""
	Get unified string diff between two strings. Trims top two lines.
	Returns empty string if strings are identical.
	"""
	_no_eol = "\ No newline at end of file"
	diffs = difflib.unified_diff(bin1.splitlines(True),bin2.splitlines(True),n=0)
	try: 
		_,_ = next(diffs),next(diffs)
	except StopIteration: 
		pass
	return ''.join([d if d[-1] == '\n' else d+'\n'+_no_eol+'\n' for d in diffs])

def apply_diff(s,patch):
	s = s.splitlines(True)
	p = patch.splitlines(True)
	t = ''
	i = sl = 0
	_hdr_pat = re.compile("^@@ -(\d+),?(\d+)? \+(\d+),?(\d+)? @@$")
	(midx,sign) = (1,'+') if not revert else (3,'-')
	while i < len(p) and p[i].startswith(("---","+++")): 
		i += 1 # skip header lines
	while i < len(p):
		m = _hdr_pat.match(p[i])
		if not m: 
			raise Exception("Bad patch -- regex mismatch [line "+str(i)+"]")
		l = int(m.group(midx))-1 + (m.group(midx+1) == '0')
		if sl > l or l > len(s):
		  raise Exception("Bad patch -- bad line num [line "+str(i)+"]")
		t += ''.join(s[sl:l])
		sl = l
		i += 1
		while i < len(p) and p[i][0] != '@':
			if i+1 < len(p) and p[i+1][0] == '\\':
				line = p[i][:-1]
				i += 2
			else: 
				line = p[i]
				i += 1
			if len(line) > 0:
				if line[0] == sign or line[0] == ' ':
					t += line[1:]
				sl += (line[0] != sign)
	t += ''.join(s[sl:])
	return t

def store_diff(storage, diff_path, key_old, key_new):
	# loads files
	# bin_old = storage.load_file_global(key_old)
	# bin_new = storage.load_file_global(key_new)

	# diff surrogate, save latest snapshot
	storage.copy_file_global(key_new,diff_path)
	return True

def update_history(init,commits):
	# loads file global
	metapath = init.prefix_meta+'history.json'
	history = json.load(init.storage.load_file_global(metapath))
	init.storage.reset_buffer()

	# adds to the latest commits
	time = datetime.now().strftime("%m/%d/%Y %H:%M:%S")
	history[str(len(history)+1)] = {'commits': commits, 'date': time}

	init.storage.add_file_from_binary_global(metapath,io.BytesIO(json.dumps(history).encode('ascii')))
	init.storage.reset_buffer()
	return True

def update_file_history(init,key,commit):
	# loads file global
	metapath = init.prefix_history+key+'/history.json'
	if init.storage.check_if_empty(init.prefix_history+key+'/'): 
		history = {}
		history[1] = commit
	else:
		history = json.load(init.storage.load_file_global(metapath))
		history[str(len(history)+1)] = commit
		init.storage.reset_buffer()
		
	init.storage.add_file_from_binary_global(metapath,io.BytesIO(json.dumps(history).encode('ascii')))
	init.storage.reset_buffer()
	return True

def add(init, files=[], location='?'):
	if location == '?':
		location = ''
	
	# adds each file
	for file in files:
		print('adding '+file+'...')
		init.storage.add_file(file, '', location)
	init.storage.reset_buffer()
	return True

def add_from_binary(init, filename='', binary='', location='?'):
	if location == '?':
		location = ''
	
	# adds each file
	print('adding '+filename+'...')
	init.storage.add_file_from_binary(filename, binary)
	init.storage.reset_buffer()
	return True

def rename_file(init, key, new_key):
	# renames the files
	init.storage.copy_file(key,new_key)
	init.storage.remove_file(key)

	# renames the diffs
	n_vers = init.get_latest_diff_number(key)
	for v in range(1,n_vers+1):
		prv_diff = init.prefix_diffs + key + '/' + str(v).zfill(10)
		nxt_diff = init.prefix_diffs + new_key + '/' + str(v).zfill(10)
		init.storage.copy_file_global(prv_diff,nxt_diff)
		init.storage.remove_file_global(prv_diff)

	prv_diff = init.prefix_history + key + '/history.json'
	nxt_diff = init.prefix_history + new_key + '/history.json'
	init.storage.copy_file_global(prv_diff,nxt_diff)

	print('rename file ' + key + ' to ' + new_key)
	return True

def copy_file_to_new_dataset(init, key, new_dataset):
	# renames the files
	init.storage.copy_file_global(init.storage.dataset+key,new_dataset+key)

	# renames the diffs
	n_vers = init.get_latest_diff_number(key)
	for v in range(1,n_vers+1):
		prv_diff = init.prefix_diffs + key + '/' + str(v).zfill(10)
		nxt_diff = init.prefix_diffs.replace(init.storage.dataset,new_dataset) + key + '/' + str(v).zfill(10)
		init.storage.copy_file_global(prv_diff,nxt_diff)
		
	prv_diff = init.prefix_history + key + '/history.json'
	nxt_diff = init.prefix_history.replace(init.storage.dataset,new_dataset) + key + '/history.json'
	init.storage.copy_file_global(prv_diff,nxt_diff)

	print('rename file ' + key + ' to ' + new_dataset)
	return True

def move_file_to_dataset(init, key, new_dataset):
	# renames the files
	init.storage.copy_file_global(init.storage.dataset+key,new_dataset+key)
	init.storage.remove_file_global(init.storage.dataset+key)

	# renames the diffs
	n_vers = init.get_latest_diff_number(key)
	for v in range(1,n_vers+1):
		prv_diff = init.prefix_diffs + key + '/' + str(v).zfill(10)
		nxt_diff = init.prefix_diffs.replace(init.storage.dataset,new_dataset) + key + '/' + str(v).zfill(10)
		init.storage.copy_file_global(prv_diff,nxt_diff)
		init.storage.remove_file_global(prv_diff)

	prv_diff = init.prefix_history + key + '/history.json'
	nxt_diff = init.prefix_history.replace(init.storage.dataset,new_dataset) + key + '/history.json'
	init.storage.copy_file_global(prv_diff,nxt_diff)
	init.storage.remove_file_global(prv_diff)

	print('moving file ' + key + ' to ' + new_dataset)
	return True

def remove(init, files=[], location='?'):
	if location == '?':
		location = ''

	# adds each file
	for file in files:
		print('deleting file '+location+file+'...')
		init.storage.remove_file(location+file)
	return True

def remove_global(init, files=[], location='?'):
	if location == '?':
		location = ''

	# adds each file
	for file in files:
		print('deleting file '+location+file+'...')
		init.storage.remove_file_global(location+file)
	return True

def remove_full(init, key):
	# diff location
	n_versions = init.get_latest_diff_number(key)
	init.storage.remove_file(key)
	for version in range(1,n_versions+1):
		diff = init.prefix_diffs + key + '/' + str(version).zfill(10)

		# generates an empty diff
		init.storage.add_file_from_binary_global(diff,io.BytesIO("".encode('ascii')))
		init.storage.reset_buffer()

	return True

def remove_diff(init, key, version):
	# diff location
	diff = init.prefix_diffs + key + '/' + str(version).zfill(10)

	# generates an empty diff
	init.storage.add_file_from_binary_global(diff,io.BytesIO("".encode('ascii')))
	init.storage.reset_buffer()
	return True

def revert_file(init, key, version):
	# finds the commit version
	diff = init.prefix_diffs + key + '/' + str(version).zfill(10)
	init.storage.copy_file_global(diff,key)

	print('reverted file ' + key + ' to version ' + str(version))
	return True

def revert_commit(init, target_version):
	if int(target_version) == 0:
		return False

	# finds the commit version
	metapath = init.prefix_meta+'history.json'
	history = json.load(init.storage.load_file_global(metapath))

	print('reverting dataset to version ' + str(target_version))

	for i in range(len(history),int(target_version),-1):
		for commit_ in history[str(i)]['commits']:
			cmit = json.load(init.storage.load_file_global(commit_))
			init.storage.reset_buffer()
			if cmit['type'] == 'add':
				remove_global(init, [cmit['key']])
			elif cmit['type'] == 'remove':
				revert_file(init,cmit['key'],cmit['version']-1)
			else:
				revert_file(init,cmit['key'],cmit['version']-1)
			init.storage.reset_buffer()
	return True

def get_key_history(init, key):
	# finds the commit version
	metapath = init.prefix_history+key+'/history.json'
	history = json.load(init.storage.load_file_global(metapath))
	return history

def pull(init, files=[],version='current'):

	if version == 'current':
		# saves each file
		for key in files:
			newFile = open(os.path.basename(key), "wb")
			binary = init.storage.load_file_global(key)
			newFile.write(binary.read())
	else:
		gtfo = False
		# finds the commit of interest
		metapath = init.prefix_meta+'history.json'
		history = json.load(init.storage.load_file_global(metapath))
		for key in files:
			if key[-1] == '/':
				print('Do not pull directories')
				return False
			for i in range(len(history),int(version)-1,-1):
				for commit in history[str(i)]['commits']:
					# reads each file version
					if init.storage.type == 'local':
						cmit = json.load(init.storage.load_file_global(commit))
					else:
						cmit = json.load(init.storage.load_file_global(commit))
					if str(cmit['version']) == version and cmit['key'] == key:
						if cmit['type'] != 'remove':
							newFile = open(os.path.basename(key), "wb")
							key = init.prefix_diffs + key + '/' + str(cmit['version']).zfill(10)
							newFile.write(init.storage.load_file_global(key).read())
							init.storage.reset_buffer()
						gtfo = True
					if gtfo:
						break
			if gtfo:
				gtfo = False
				break

	init.storage.reset_buffer()

	return True

def print_history(init):
	# loads file global
	metapath = init.prefix_meta+'history.json'
	history = json.load(init.storage.load_file_global(metapath))
	init.storage.reset_buffer()

	print('\n--------------------------')
	print('- History of the dataset -')
	print('--------------------------\n')

	# prints history
	for i in range(len(history),0,-1):
		print('Version: '+str(i)+' Date: '+ history[str(i)]['date'])
		idx = 0
		for commit in history[str(i)]['commits']:
			# reads each file version
			if init.storage.type == 'local':
				cmit = json.load(init.storage.load_file_global(commit))
			else:
				cmit = json.load(init.storage.load_file_global(commit))
			print('-- '+cmit['comment']+' by '+cmit['source'])
			idx += 1
			if idx > 4:
				print('-- ('+str(len(history[str(i)]['commits'])-idx)+' more changes not showed)')
				break

	print('')

	init.storage.reset_buffer()

def print_diff(init, v2, v1, file=''):
	# loads file global
	metapath = init.prefix_meta+'history.json'
	history = json.load(init.storage.load_file_global(metapath))
	init.storage.reset_buffer()

	v2 = int(v2)
	v1 = int(v1) 

	if file == '':
		# prints history
		print('Comparing commit '+str(v2)+' and commit '+str(v1))
		print('Commit: '+str(v2)+' Date: '+ history[str(v2)]['date'])
		print('Commit: '+str(v1)+' Date: '+ history[str(max(v1,1))]['date'])

		for i in range(v2,v1,-1):
			for commit in history[str(i)]['commits']:
				# reads each file version
				if init.storage.type == 'local':
					cmit = json.load(init.storage.load_file_global(commit))
				else:
					cmit = json.load(init.storage.load_file_global(commit))
				print('-- '+cmit['comment']+' by '+cmit['source'])
	else:
		print('Comparing file '+file+' between commit '+str(v2)+' and commit '+str(v1))
		idx = 0
		# prints history
		for i in range(v2,v1,-1):
			for commit in history[str(i)]['commits']:
				# reads each file version
				if init.storage.type == 'local':
					cmit = json.load(init.storage.load_file_global(commit))
				else:
					cmit = json.load(init.storage.load_file_global(commit))
				# print(cmit['comment']+' by '+cmit['source'])
				if cmit['key'] == init.storage.dataset+file:
					print('Version: '+str(cmit['version'])+' Date: '+ history[str(i)]['date'])
					print('-- '+cmit['comment']+' by '+cmit['source'])
				if str(cmit['version']) == '1':
					init.storage.reset_buffer()
					return True
					break
		init.storage.reset_buffer()